from schematics.models import Model
from schematics.types import ListType, StringType


class CMDCommandExample(Model):
    name = StringType(required=True)
    commands = ListType(StringType(), required=True, min_size=1)





class PlaneEnum:
    Mgmt = "mgmt-plane"

    # # data planes should be subdivided by module
    # Data_KeyVault = "data-plane_keyvault"

    _config = {
        Mgmt: {
            "client": "MgmtClient",  # MgmtClient is implemented in azure.cli.core.aaz._client
        },
        # Data_KeyVault: {
        #     "swagger_modules": ("keyvault", ),
        # }
    }

    @classmethod
    def choices(cls):
        return tuple(v for k, v in vars(cls).items() if not k.startswith('_') and isinstance(v, str))

    @classmethod
    def is_valid_swagger_module(cls, plane, module_name):
        swagger_modules = cls._config[plane].get('swagger_modules', None)
        if not swagger_modules or module_name in swagger_modules:
            return True
        return False

    @classmethod
    def http_client(cls, plane):
        return cls._config[plane]['client']


__all__ = ["PlaneEnum"]



import json
import logging
import os
from collections import OrderedDict

from swagger.utils import exceptions

logger = logging.getLogger('backend')


class SwaggerLoader:

    def __init__(self):
        self._loaded = {}
        self.loaded_swaggers = OrderedDict()
        self._linked_idx = 0

    def load_file(self, file_path):
        from swagger.model.schema.swagger import Swagger
        loaded = self.get_loaded(file_path)
        if loaded is not None:
            return loaded

        with open(file_path, 'r', encoding='utf-8') as f:
            body = json.load(f)

        if 'example' in file_path.lower():
            loaded = body
        else:
            self.patch_swagger(body)
            loaded = Swagger(body)
            self.loaded_swaggers[file_path] = loaded
        self._cache_loaded(loaded, file_path)
        return loaded

    @staticmethod
    def patch_swagger(body):
        """Current will patch `additionalProperties: {}` expression"""
        def _patch_list(lst):
            for item in lst:
                if isinstance(item, dict):
                    _patch_dict(item)

        def _patch_dict(dct):
            for key in [*dct.keys()]:
                if key == "additionalProperties" and dct[key] == {}:
                    # replace `additionalProperties: {}` by `additionalProperties: true`
                    dct[key] = True
                _patch(dct[key])

        def _patch(data):
            if isinstance(data, dict):
                _patch_dict(data)
            elif isinstance(data, list):
                _patch_list(data)

        _patch(body)

    def link_swaggers(self):
        while self._linked_idx < len(self.loaded_swaggers):
            file_path, swagger = [*self.loaded_swaggers.items()][self._linked_idx]
            swagger.link(self, file_path)
            self._linked_idx += 1

    def get_loaded(self, *traces):
        return self._loaded.get(traces, None)

    def _cache_loaded(self, loaded, *traces):
        self._loaded[traces] = loaded

    def load_ref(self, ref_link, *ref_traces):
        traces = self._parse_ref_link(ref_traces, ref_link)

        ref = self.get_loaded(*traces)
        if ref is not None:
            return ref, traces

        file_path = traces[0]
        ref = self.get_loaded(file_path)
        if ref is None:
            try:
                ref = self.load_file(file_path)
            except FileNotFoundError:
                raise exceptions.InvalidSwaggerValueError(
                    msg='Cannot find reference swagger file',
                    key=ref_traces, value=ref_link)

        for prop in traces[1:]:
            assert prop != ''
            try:
                if isinstance(ref, dict):
                    ref = ref[prop]
                elif isinstance(ref, list):
                    ref = ref[prop]
                else:
                    ref = getattr(ref, prop)
            except (KeyError, AttributeError):
                raise exceptions.InvalidSwaggerValueError(
                    msg='Failed to find reference in swagger',
                    key=ref_traces, value=ref_link)

        assert ref is not None
        self._cache_loaded(ref, *traces)
        return ref, traces

    @classmethod
    def _parse_ref_link(cls, ref_traces, ref_link):
        file_path = ref_traces[0]

        parts = ref_link.strip().split('#')

        if not (1 <= len(parts) <= 2):
            raise exceptions.InvalidSwaggerValueError(
                msg="Invalid Reference Value",
                key=ref_traces, value=ref_link)

        if parts[0] == '':
            # start with '#'
            traces = [trace for trace in parts[1].split('/') if trace]
            return file_path, *traces

        # find ref_file_path
        file_path = os.path.dirname(file_path)
        for s in parts[0].split('/'):
            if s == '.':
                continue
            elif s == '':
                continue
            elif s == '..':
                file_path = os.path.dirname(file_path)
            else:
                file_path = os.path.join(file_path, s)
        if len(parts) == 2:
            traces = [trace for trace in parts[1].split('/') if trace]
            return file_path, *traces
        else:
            return file_path,



from schematics.models import Model
from schematics.types import StringType, ListType


class CMDHelp(Model):
    # properties as tags
    short = StringType(required=True)  # short-summary

    # properties as nodes
    lines = ListType(StringType())  # long-summary separated by lines

    class Options:
        serialize_when_none = False


class CMDArgumentHelp(CMDHelp):
    # properties as nodes
    ref_commands = ListType(
        StringType(),
        serialized_name='refCommands',
        deserialize_from='refCommands',
    )  # popular commands



import re


def to_camel_case(name):
    assert isinstance(name, str)
    parts = name.replace('-', ' ').replace('_', ' ').split()
    parts = [p[0].upper() + p[1:] for p in parts if p]
    return "".join(parts)


def to_snake_case(name, separator='_'):
    assert isinstance(name, str)
    name = re.sub('(.)([A-Z][a-z]+)', r'\1' + separator + r'\2', name)
    name = re.sub('([a-z0-9])([A-Z])', r'\1' + separator + r'\2', name).lower()
    return name.replace('-', separator).replace('_', separator)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from schematics.models import Model
from schematics.types import StringType, ModelType, ListType, PolyModelType, IntType

from ._fields import CMDVariantField, CMDBooleanField, CMDURLPathField, CMDDescriptionField
from ._http_request_body import CMDHttpRequestBody
from ._http_response_body import CMDHttpResponseBody
from ._schema import CMDSchemaField
from ._arg_builder import CMDArgBuilder
from ._arg import CMDResourceGroupNameArg, CMDSubscriptionIdArg, CMDResourceLocationArg
from ._utils import CMDDiffLevelEnum
from msrestazure.tools import parse_resource_id, resource_id
from ._utils import CMDArgBuildPrefix


class CMDHttpRequestArgs(Model):
    # properties as nodes
    params = ListType(CMDSchemaField())
    consts = ListType(CMDSchemaField())

    class Options:
        serialize_when_none = False

    def reformat(self, **kwargs):
        if self.params:
            for param in self.params:
                param.reformat(**kwargs)
            self.params = sorted(self.params, key=lambda s: s.name)
        if self.consts:
            for const in self.consts:
                const.reformat(**kwargs)
            self.consts = sorted(self.consts, key=lambda c: c.name)


class CMDHttpRequestPath(CMDHttpRequestArgs):

    def generate_args(self, path, ref_args, has_subresource):
        try:
            id_parts = parse_resource_id(path)
            resource_id(**id_parts)
        except KeyError:
            # not a valid resource id
            id_parts = {}
        resource_name = id_parts.pop("resource_name", None)
        if resource_name and (not path.endswith(resource_name) or has_subresource):
            resource_name = None
        args = []
        if self.params:
            var_prefix = CMDArgBuildPrefix.Path
            for param in self.params:
                id_part = None
                placeholder = '{' + param.name + '}'
                for part, name in id_parts.items():
                    if name == placeholder:
                        id_part = part
                        break

                if id_part == 'subscription' or param.name == 'subscriptionId':
                    arg = CMDSubscriptionIdArg({
                        'var': f'{var_prefix}.subscriptionId',
                        'options': ['subscription'],
                    })
                    param.arg = arg.var
                    arg.ref_schema = param
                elif id_part == 'resource_group' or param.name == 'resourceGroupName':
                    arg = CMDResourceGroupNameArg({
                        'var': f'{var_prefix}.resourceGroupName',
                        'options': ['resource-group', 'g'],
                    })
                    param.arg = arg.var
                    arg.ref_schema = param
                elif param.name == 'location':
                    arg = CMDResourceLocationArg({
                        'var': f'{var_prefix}.location',
                        'options': ['location', 'l'],
                    })
                    param.arg = arg.var
                    arg.ref_schema = param
                else:
                    builder = CMDArgBuilder.new_builder(
                        schema=param, var_prefix=var_prefix, ref_args=ref_args
                    )
                    result = builder.get_args()
                    assert len(result) == 1
                    arg = result[0]

                    if resource_name == placeholder and not builder._ref_arg:
                        arg.options = list({*arg.options, "name", "n"})

                arg.required = True
                arg.id_part = id_part  # id_part should not be generated for create command or commands for subresource
                args.append(arg)

        return args


class CMDHttpRequestQuery(CMDHttpRequestArgs):

    def generate_args(self, ref_args):
        args = []
        if self.params:
            for param in self.params:
                builder = CMDArgBuilder.new_builder(
                    schema=param, var_prefix=CMDArgBuildPrefix.Query, ref_args=ref_args
                )
                args.extend(builder.get_args())
        return args


class CMDHttpRequestHeader(CMDHttpRequestArgs):
    # properties as tags
    client_request_id = StringType(
        serialized_name="clientRequestId",
        deserialize_from="clientRequestId",
    )  # specifies the header parameter to be used instead of `x-ms-client-request-id`

    def generate_args(self, ref_args):
        args = []
        for param in self.params:
            builder = CMDArgBuilder.new_builder(
                schema=param, var_prefix=CMDArgBuildPrefix.Header, ref_args=ref_args
            )
            args.extend(builder.get_args())
        return args


class CMDHttpRequest(Model):
    # properties as tags
    method = StringType(choices=("get", "put", "post", "delete", "options", "head", "patch",), required=True)

    # properties as nodes
    path = ModelType(CMDHttpRequestPath)
    query = ModelType(CMDHttpRequestQuery)
    header = ModelType(CMDHttpRequestHeader)
    body = PolyModelType(CMDHttpRequestBody, allow_subclasses=True)

    class Options:
        serialize_when_none = False

    def generate_args(self, path, ref_args, has_subresource):
        args = []
        if self.path:
            args.extend(self.path.generate_args(path, ref_args, has_subresource))
        if self.query:
            args.extend(self.query.generate_args(ref_args))
        if self.header:
            args.extend(self.header.generate_args(ref_args))
        if self.body:
            args.extend(self.body.generate_args(ref_args))
        return args

    def reformat(self, **kwargs):
        if self.path:
            self.path.reformat(**kwargs)
        if self.query:
            self.query.reformat(**kwargs)
        if self.header:
            self.header.reformat(**kwargs)
        if self.body:
            self.body.reformat(**kwargs)

    def register_cls(self, **kwargs):
        if self.body:
            self.body.register_cls(**kwargs)


class CMDHttpResponseHeaderItem(Model):
    # properties as tags
    name = StringType(required=True)
    var = CMDVariantField()

    class Options:
        serialize_when_none = False

    def diff(self, old, level):
        diff = {}

        if level >= CMDDiffLevelEnum.Structure:
            if old.var and not self.var:
                diff["var"] = "Miss variant"
            if self.var and not old.var:
                diff["var"] = "New variant"

        if level >= CMDDiffLevelEnum.Associate:
            if self.var != old.var:
                diff["var"] = f"{old.var} != {self.var}"
        return diff


class CMDHttpResponseHeader(Model):
    # properties as nodes
    items = ListType(ModelType(CMDHttpResponseHeaderItem))

    class Options:
        serialize_when_none = False

    def diff(self, old, level):
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            items_dict = {item.name: item for item in self.items}
            for old_item in old.items:
                if old_item.name not in items_dict:
                    diff[old_item.name] = "Miss header item"
                else:
                    item = items_dict.pop(old_item.name)
                    item_diff = item.diff(old_item, level)
                    if item_diff:
                        diff[old_item.name] = item_diff

        if level >= CMDDiffLevelEnum.Structure:
            old_items_dict = {item.name: item for item in old.items}
            for item in self.items:
                if item.name not in old_items_dict:
                    diff[item.name] = "New header item"
                else:
                    old_items_dict.pop(item.name)
            for old_item in old_items_dict.values():
                diff[old_item.name] = "Miss header item"
        return diff

    def reformat(self, **kwargs):
        if self.items:
            self.items = sorted(self.items, key=lambda h: h.name)


class CMDHttpResponse(Model):
    # properties as tags
    status_codes = ListType(
        IntType(),
        serialized_name='statusCode',
        deserialize_from='statusCode',
    )  # if status_codes is None, then it's the default response.
    is_error = CMDBooleanField(
        serialized_name='isError',
        deserialize_from='isError'
    )
    description = CMDDescriptionField()

    # properties as nodes
    header = ModelType(CMDHttpResponseHeader)
    body = PolyModelType(CMDHttpResponseBody, allow_subclasses=True)

    class Options:
        serialize_when_none = False

    def diff(self, old, level):
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if (self.status_codes is not None) != (old.status_codes is not None):
                diff["status_codes"] = f"{old.status_codes} != {self.status_codes}"
            elif self.status_codes and set(self.status_codes) != set(old.status_codes):
                diff["status_codes"] = f"{old.status_codes} != {self.status_codes}"

            if (not self.is_error) != (not old.is_error):
                diff["is_error"] = f"{old.is_error} != {self.is_error}"

            if old.header is not None:
                if self.header is None:
                    diff["header"] = "Miss header"
                else:
                    header_diff = self.header.diff(old.header, level)
                    if header_diff:
                        diff["header"] = header_diff

            if old.body is not None:
                if self.body is None:
                    diff["body"] = "Miss response body"
                else:
                    body_diff = self.body.diff(old.body, level)
                    if body_diff:
                        diff["body"] = body_diff

        if level >= CMDDiffLevelEnum.Structure:
            if self.header is not None and old.header is None:
                diff["header"] = "New header"
            if self.body is not None and old.body is None:
                diff["body"] = "New response body"

        if level >= CMDDiffLevelEnum.All:
            if self.description != old.description:
                diff["description"] = f"'{old.description}' != '{self.description}'"

        return diff

    def reformat(self, **kwargs):
        if self.header:
            self.header.reformat(**kwargs)
        if self.body:
            self.body.reformat(**kwargs)

    def register_cls(self, **kwargs):
        if self.body:
            self.body.register_cls(**kwargs)


class CMDHttpAction(Model):
    # properties as tags
    path = CMDURLPathField(required=True)

    # properties as nodes
    request = ModelType(CMDHttpRequest)
    responses = ListType(ModelType(CMDHttpResponse))

    def generate_args(self, ref_args, has_subresource):
        return self.request.generate_args(path=self.path, ref_args=ref_args, has_subresource=has_subresource)

    def reformat(self, **kwargs):
        if self.request:
            self.request.reformat(**kwargs)
        if self.responses:
            for response in self.responses:
                if response.is_error:
                    continue
                response.reformat(**kwargs)

    def register_cls(self, **kwargs):
        if self.request:
            self.request.register_cls(**kwargs)
        if self.responses:
            for response in self.responses:
                if response.is_error:
                    continue
                response.register_cls(**kwargs)



from schematics.models import Model
from schematics.types import StringType
from uuid import uuid4


class Linkable:

    def __init__(self):
        self._linked = False
        self.traces = None
        self._random_uuid = uuid4()

    def is_linked(self):
        if not hasattr(self, '_linked'):
            self._linked = False
        return self._linked

    def link(self, swagger_loader, *traces, **kwargs):
        self._linked = True
        assert len(traces) > 0
        self.traces = tuple(traces)
        for trace in self.traces:
            assert isinstance(trace, (str, int))

    def __hash__(self):
        return hash(self._random_uuid)


class ReferenceField(StringType):

    def __init__(self, serialized_name="$ref", deserialize_from="$ref", **kwargs):

        super(ReferenceField, self).__init__(
            serialized_name=serialized_name,
            deserialize_from=deserialize_from,
            **kwargs
        )


class Reference(Model, Linkable):
    """A simple object to allow referencing other definitions in the specification. It can be used to reference parameters and responses that are defined at the top level for reuse."""

    ref = ReferenceField(required=True)  # The reference string

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ref_instance = None

    def link(self, swagger_loader, *traces, **kwargs):
        if self.is_linked():
            return
        super().link(swagger_loader, *traces, **kwargs)

        self.ref_instance, instance_traces = swagger_loader.load_ref(self.ref, *self.traces, 'ref')
        if isinstance(self.ref_instance, Linkable):
            self.ref_instance.link(swagger_loader, *instance_traces, **kwargs)

    @classmethod
    def _claim_polymorphic(cls, data):
        return isinstance(data, dict) and "$ref" in data and len(data) == 1



from flask import Blueprint, jsonify, request, url_for

from utils.config import Config
from utils import exceptions
from cli.controller.az_module_manager import AzMainManager, AzExtensionManager
from cli.controller.portal_cli_generator import PortalCliGenerator
from cli.model.view import CLIModule
from command.controller.specs_manager import AAZSpecsManager
import logging

logging.basicConfig(level="INFO")


bp = Blueprint('az', __name__, url_prefix='/CLI/Az')


@bp.route("/Profiles", methods=("GET", ))
def az_profiles():
    return jsonify(Config.CLI_PROFILES)


@bp.route("/Main/Modules", methods=("GET", "POST"))
def az_main_modules():
    manager = AzMainManager()
    if request.method == "POST":
        # create a new module in azure-cli
        data = request.get_json()
        if not data or not isinstance(data, dict) or 'name' not in data:
            raise exceptions.InvalidAPIUsage("Invalid request body")
        mod_name = data['name']
        module = manager.create_new_mod(mod_name)
        result = module.to_primitive()
        result['url'] = url_for('az.az_main_module', module_name=module.name)
    elif request.method == "GET":
        # list available modules
        modules = manager.list_modules()
        result = []
        for module in modules:
            result.append({
                **module,
                'url': url_for('az.az_main_module', module_name=module['name']),
            })
    else:
        raise NotImplementedError()
    return jsonify(result)


@bp.route("/Main/Modules/<Name:module_name>", methods=("GET", "PUT", "PATCH"))
def az_main_module(module_name):
    manager = AzMainManager()
    if request.method == "PUT":
        data = request.get_json()
        if 'profiles' not in data:
            raise exceptions.InvalidAPIUsage("miss profiles for module")
        module = CLIModule({
            "name": module_name,
            "profiles": data['profiles']
        })
        module = manager.update_module(module_name, module.profiles)
        result = module.to_primitive()
        result['url'] = url_for('az.az_main_module', module_name=module.name)
    elif request.method == "PATCH":
        data = request.get_json()
        if 'profiles' not in data:
            raise exceptions.InvalidAPIUsage("miss profiles for module")
        module = CLIModule({
            "name": module_name,
            "profiles": data['profiles']
        })
        module = manager.update_module(module_name, module.profiles, by_patch=True)
        result = module.to_primitive()
        result['url'] = url_for('az.az_main_module', module_name=module.name)
    elif request.method == "GET":
        module = manager.load_module(module_name)
        result = module.to_primitive()
        result['url'] = url_for('az.az_main_module', module_name=module.name)
    else:
        raise NotImplementedError()
    return jsonify(result)


@bp.route("/Extension/Modules", methods=("GET", "POST"))
def az_extension_modules():
    manager = AzExtensionManager()
    if request.method == "POST":
        # create a new extension in azure-cli-extensions
        data = request.get_json()
        if not data or not isinstance(data, dict) or 'name' not in data:
            raise exceptions.InvalidAPIUsage("Invalid request body")
        mod_name = data['name']
        module = manager.create_new_mod(mod_name)
        result = module.to_primitive()
        result['url'] = url_for('az.az_extension_module', module_name=module.name)
    elif request.method == "GET":
        # list available modules
        modules = manager.list_modules()
        result = []
        for module in modules:
            result.append({
                **module,
                'url': url_for('az.az_extension_module', module_name=module['name']),
            })
    else:
        raise NotImplementedError()
    return jsonify(result)


@bp.route("/Extension/Modules/<Name:module_name>", methods=("GET", "PUT", "PATCH"))
def az_extension_module(module_name):
    manager = AzExtensionManager()
    if request.method == "PUT":
        data = request.get_json()
        if 'profiles' not in data:
            raise exceptions.InvalidAPIUsage("miss profiles for module")
        module = CLIModule({
            "name": module_name,
            "profiles": data['profiles']
        })
        module = manager.update_module(module_name, module.profiles)
        result = module.to_primitive()
        result['url'] = url_for('az.az_extension_module', module_name=module.name)
    elif request.method == "PATCH":
        data = request.get_json()
        if 'profiles' not in data:
            raise exceptions.InvalidAPIUsage("miss profiles for module")
        module = CLIModule({
            "name": module_name,
            "profiles": data['profiles']
        })
        module = manager.update_module(module_name, module.profiles, by_patch=True)
        result = module.to_primitive()
        result['url'] = url_for('az.az_extension_module', module_name=module.name)
    elif request.method == "GET":
        module = manager.load_module(module_name)
        result = module.to_primitive()
        result['url'] = url_for('az.az_extension_module', module_name=module.name)
    else:
        raise NotImplementedError()
    return jsonify(result)


@bp.route("/Main/Modules/<Name:module_name>/ExportPortalConfig", methods=("POST",))
def portal_generate_main_module(module_name):
    az_main_manager = AzMainManager()
    if az_main_manager.has_module(module_name):
        cli_module = az_main_manager.load_module(module_name)
        registered_cmds = az_main_manager.find_module_cmd_registered(cli_module['profiles']['latest'])
        logging.info("Input module: {0} in cli".format(module_name))
    else:
        exceptions.ResourceNotFind("Invalid input module: {0}, please check".format(module_name))
        return

    aaz_spec_manager = AAZSpecsManager()
    root = aaz_spec_manager.find_command_group()
    if not root:
        raise exceptions.ResourceNotFind("Command group not exist")
    portal_cli_generator = PortalCliGenerator()
    cmd_portal_list = portal_cli_generator.generate_cmds_portal_info(aaz_spec_manager, registered_cmds)
    portal_cli_generator.generate_cmds_portal(cmd_portal_list)
    return "Done"


@bp.route("/Extension/Modules/<Name:module_name>/ExportPortalConfig", methods=("POST",))
def portal_generate_extension_module(module_name):
    az_ext_manager = AzExtensionManager()
    if az_ext_manager.has_module(module_name):
        cli_module = az_ext_manager.load_module(module_name)
        registered_cmds = az_ext_manager.find_module_cmd_registered(cli_module['profiles']['latest'])
        logging.info("Input module: {0} in cli extension".format(module_name))
    else:
        exceptions.ResourceNotFind("Invalid input module: {0}, please check".format(module_name))
        return

    aaz_spec_manager = AAZSpecsManager()
    root = aaz_spec_manager.find_command_group()
    if not root:
        raise exceptions.ResourceNotFind("Command group not exist")
    portal_cli_generator = PortalCliGenerator()
    cmd_portal_list = portal_cli_generator.generate_cmds_portal_info(aaz_spec_manager, registered_cmds)
    portal_cli_generator.generate_cmds_portal(cmd_portal_list)
    return "Done"




_MAJOR, _MINOR, _PATCH, _SUFFIX = ("1", "7", "1", "")

# _PATCH: On main and in a nightly release the patch should be one ahead of the last released build.
# _SUFFIX: This is mainly for nightly builds which have the suffix ".dev$DATE". See
# https://semver.org/#is-v123-a-semantic-version for the semantics.

VERSION_SHORT = f"{_MAJOR}.{_MINOR}"
VERSION = f"{_MAJOR}.{_MINOR}.{_PATCH}{_SUFFIX}"



from schematics.types import StringType


class AAZStageEnum:
    Experimental = "Experimental"
    Preview = "Preview"
    Stable = "Stable"


class AAZStageField(StringType):
    """The stage for command group, command or argument."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            choices=(AAZStageEnum.Experimental, AAZStageEnum.Preview, AAZStageEnum.Stable),
            default=AAZStageEnum.Stable,
            serialize_when_none=False,
            *args,
            **kwargs
        )



import json


class InvalidAPIUsage(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        super().__init__()
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv

    def __str__(self):
        s = f"{self.message}"
        if self.status_code:
            s +=f'\nstatus_code: {self.status_code}'
        if self.payload:
            try:
                s += f'\npayload: {json.dumps(self.payload)}'
            except:
                s += f'\npayload: {self.payload}'
        return s


class VerificationError(InvalidAPIUsage):

    def __init__(self, message, details, status_code=None):
        super().__init__(message=message, status_code=status_code, payload={
            'details': details
        })


class ResourceNotFind(InvalidAPIUsage):
    status_code = 404


class ResourceConflict(InvalidAPIUsage):
    status_code = 409





from jinja2.filters import pass_environment
from utils.base64 import b64encode_str
from utils.stage import AAZStageEnum


@pass_environment
def command_group_readme_path(env, names):
    return '/'.join(["", "Commands", *names, 'readme.md'])


@pass_environment
def command_readme_path(env, names):
    return '/'.join(["", "Commands", *names[:-1], f'_{names[-1]}.md'])


@pass_environment
def resource_cfg_path(env, resource):
    return '/'.join(["", "Resources", resource.plane, b64encode_str(resource.id), f"{resource.version}.xml"])


@pass_environment
def stage_label(env, stage, bold=True):
    text = stage
    if stage is None:
        text = AAZStageEnum.Stable
    if bold:
        text = f"**{text}**"
    return text


custom_filters = {
    "command_group_readme_path": command_group_readme_path,
    "command_readme_path": command_readme_path,
    "resource_cfg_path": resource_cfg_path,
    "stage_label": stage_label
}



from command.model.configuration import CMDCommand, CMDHttpOperation
from command.model.specs import CMDSpecsCommand, CMDSpecsCommandVersion
from utils import exceptions, portal_file_schema
from utils.config import Config
import os, json, re, logging
import jsonschema

class PortalCliGenerator:
    COMMAND_ROOT_NAME = "az"
    DOC_ROOT_NAME = "https://docs.microsoft.com/cli/azure"

    # az change-analysis list --star-ttime bbb --endtime aaa -g aaa
    PARA_REG_PATTERN = re.compile(r" --?([a-zA-Z0-9\-]+) ?")

    # /subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}...
    # /providers/Microsoft.Network/dnsResolvers/{dnsResolverName}/inboundEndpoints/{inboundEndpointName}
    RESOURCE_PATH_PATTERN = re.compile(r"/([a-zA-Z0-9\-]+)/?")

    # /subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}...
    # /providers/Microsoft.OperationalInsights/workspaces/{workspaceName}...
    # /providers/Microsoft.SecurityInsights/incidents/{incidentId}

    def __init__(self):
        pass

    def fill_resource_type(self, cmd_portal_info, cmd_cfg, leaf):
        """
        path: ....Microsoft.Network/dnsForwardingRulesets/{dnsForwardingRulesetName}/forwardingRules/{forwardingRuleName}
        resource_type: dnsForwardingRulesets/forwardingRules
        path: ....Microsoft.Network/dnsForwardingRulesets/{dnsForwardingRulesetName}/forwardingRules -> list cmd
        resource_type: dnsForwardingRulesets
        """
        if cmd_cfg.resources is None:
            raise exceptions.ResourceNotFind("Resource not exist")
        resource = cmd_cfg.resources[-1]
        if resource.rp_name is None:
            raise exceptions.ResourceNotFind("Resource rp_name not exist")
        if resource.swagger_path is None:
            raise exceptions.ResourceNotFind("Resource swagger_path not exist")
        rp_name = resource.rp_name
        swagger_path = resource.swagger_path
        cmd_portal_info['rp_name'] = rp_name
        swagger_provider_ind = swagger_path.lower().rfind(rp_name.lower())
        if swagger_provider_ind == -1:
            logging.warning("please check cmd " + " ".join(leaf.names) + " resource id: " + swagger_path)
            return
        folder_end_ind = swagger_path.rfind("{", swagger_provider_ind)
        resource_paths = re.finditer(self.RESOURCE_PATH_PATTERN, swagger_path[swagger_provider_ind:folder_end_ind])
        resource_type = ""
        for resource_match in resource_paths:
            resource_folder = resource_match.group(1)
            if len(resource_type) == 0:
                resource_type += resource_folder
            else:
                resource_type += "/" + resource_folder
        cmd_portal_info['resourceType'] = resource_type

    def fill_api_version(self, cmd_portal_info, target_version):
        if target_version.name is None:
            raise exceptions.ResourceNotFind("version name not exist")
        cmd_portal_info['apiVersion'] = target_version.name

    def generate_base_learn_more(self, cmd_portal_info, leaf):
        """
        cmd: az xx yy zz
        leaf.names: [xx, yy, zz]
        help link: https://learn.microsoft.com/cli/azure/xx/yy
        """
        base_help_url = self.DOC_ROOT_NAME + "/" + "/".join(leaf.names[:-1])
        cmd_portal_info['learnMore'] = {'url': base_help_url}

    def generate_cmd_learn_more(self, help_info, leaf):
        """
        cmd name: az xx yy zz
        leaf.names: [xx, yy, zz]
        help link: https://learn.microsoft.com/cli/azure/xx/yy#az-xx-yy-zz
        """
        cmd_help_url = self.DOC_ROOT_NAME + "/" + "/".join(leaf.names[:-1]) + "#" + \
                       "-".join([self.COMMAND_ROOT_NAME] + leaf.names)
        help_info['learnMore'] = {'url' : cmd_help_url}

    def fill_cmd_name(self, cmd_info, leaf):
        cmd_name = " ".join([self.COMMAND_ROOT_NAME] + leaf.names)
        cmd_info['name'] = cmd_name

    def fill_cmd_desc(self, cmd_info, leaf):
        cmd_info['description'] = leaf.help.to_primitive()['short']

    def fill_cmd_path(self, cmd_info, cmd_cfg):
        if cmd_cfg.resources is None:
            raise exceptions.ResourceNotFind("Resource not exist")
        resource = cmd_cfg.resources[-1]
        cmd_info['path'] = resource.swagger_path

    def fill_cmd_confirmation(self, cmd_info, cmd_cfg):
        if cmd_cfg.confirmation:
            cmd_info['confirmation'] = True
        else:
            cmd_info['confirmation'] = False

    def fill_cmd_help(self, cmd_info, cmd_cfg, leaf):
        help_info = {}
        self.generate_cmd_learn_more(help_info, leaf)
        option_list = []
        for arg_group in cmd_cfg.arg_groups:
            for arg in arg_group.args:
                assert arg.options is not None
                if len(arg.options) > 0:
                    # parameter set ignore 1 character option
                    option_list += filter(lambda opt: len(opt) > 1, arg.options)
        parameters_list = [ "[--" + param_name + "]" for param_name in option_list]
        help_info['parameterSets'] = [{'parameters': parameters_list}]
        cmd_info['help'] = help_info

    def generate_default_example(self, cmd_info, leaf, var_option_list):
        parameters = []
        for (para_var, raw_options, is_required, is_in_path) in var_option_list:
            if not is_required:
                continue
            pick_options = list(filter(lambda opt: len(opt) > 1, raw_options))
            pick_option = pick_options[0]
            parameters.append({
                'name': ("-" if len(pick_option) == 1 else "--") + pick_option,
                'value': "[" + para_var.replace("$Path", "path") + "]"
            })
        if len(parameters) > 0:
            cmd_info['examples'] = [{
                'description': leaf.help.to_primitive()['short'],
                'parameters': parameters
            }]

    def find_para_var(self, para, var_option_list):
        for (var, options, is_required, is_in_path) in var_option_list:
            if para in options:
                return (var, options, is_required, is_in_path)
        return (None, None, None, None)

    def fill_cmd_examples(self, cmd_info, cmd_cfg, leaf, target_version):
        var_option_list = []
        var_required = set()
        sub_id_required = False
        for arg_group in cmd_cfg.arg_groups:
            for arg in arg_group.args:
                assert arg.options is not None
                if len(arg.options) == 0: continue
                is_required = arg.required
                var_name = arg.var
                if arg.required and var_name.find("subscriptionId") == -1:
                    # for user input cmd example, ignore subscriptionId
                    var_required.add(var_name)
                if arg.required and var_name == "$Path.subscriptionId":
                    sub_id_required = True
                if cmd_info['path'].find("/resourceGroups/{") != -1 and var_name == "$Path.resourceGroupName":
                    var_required.add(var_name)
                    is_required = True
                var_option_list.append((var_name, arg.options,
                                        is_required,
                                        var_name.find('$Path') != -1))

        if not target_version.examples:
            self.generate_default_example(cmd_info, leaf, var_option_list)
            return
        find_example = False
        for example in target_version.examples:
            if find_example: break
            assert example.name is not None
            assert example.commands is not None
            for command_str in example.commands:
                if find_example: break
                param_verify = True
                para_list = []
                para_groups = re.finditer(self.PARA_REG_PATTERN, command_str)
                if not para_groups:
                    continue
                cmd_para = set()
                # 1. check if all parameters in example command str is in path, if not ignore
                for para_match in para_groups:
                    para_name = para_match.group(1)
                    var, raw_options, is_required, is_in_path = self.find_para_var(para_name, var_option_list)
                    # var not found, drop this command_str
                    if not var or not is_in_path:
                        param_verify = False
                        break
                    pick_options = list(filter(lambda opt: len(opt) > 1, raw_options))
                    para_list.append((var, pick_options[0]))
                    cmd_para.add(var)

                # 2. check if all required parameters is in example command str
                # if not, drop this example command str
                if not var_required.issubset(cmd_para):
                    param_verify = False
                if not param_verify: continue
                parameters = []
                sub_id_added = False
                for (var, pick_option) in para_list:
                    if var == "$Path.subscriptionId":
                        sub_id_added = True
                    parameters.append({
                        'name': "--" + pick_option,
                        'value': "[" + var.replace("$Path", "path") + "]"
                    })
                if not sub_id_added and sub_id_required:
                    parameters.append({
                        'name': "--subscription",
                        'value': "[path.subscriptionId]"
                    })

                cmd_info['examples'] = [{
                    'description': example.name,
                    'parameters': parameters
                }]
                find_example = True
        if not find_example:
            self.generate_default_example(cmd_info, leaf, var_option_list)

    def __check_cmd_examples(self, cmd_cfg):
        parameter_verified = True
        for arg_group in cmd_cfg.arg_groups:
            for arg in arg_group.args:
                assert arg.options is not None
                if len(arg.options) == 0: continue
                var_name = arg.var
                if arg.required and var_name.find('$Path') == -1:
                    parameter_verified = False
                    break
        if cmd_cfg.operations:
            for opt in cmd_cfg.operations:
                if isinstance(opt, CMDHttpOperation) \
                        and opt.http.request is not None \
                        and opt.http.request.body is not None:
                    parameter_verified = False
                    break
        return parameter_verified

    def generate_cmd_info(self, cmd_info, cmd_cfg, leaf, target_version):
        self.fill_cmd_name(cmd_info, leaf)
        self.fill_cmd_desc(cmd_info, leaf)
        self.fill_cmd_path(cmd_info, cmd_cfg)
        self.fill_cmd_confirmation(cmd_info, cmd_cfg)
        self.fill_cmd_help(cmd_info, cmd_cfg, leaf)
        if self.__check_cmd_examples(cmd_cfg):
            self.fill_cmd_examples(cmd_info, cmd_cfg, leaf, target_version)

    def generator_command_portal(self, cmd_cfg, leaf, target_version):
        if not isinstance(cmd_cfg, CMDCommand) or \
                not isinstance(leaf, CMDSpecsCommand) or \
                not isinstance(target_version, CMDSpecsCommandVersion):
            return None
        if cmd_cfg.arg_groups is None:
            logging.info("{0} cmd has no arg_group".format(" ".join(leaf.names)))
            return None
        cmd_portal_info = {}
        self.fill_resource_type(cmd_portal_info, cmd_cfg, leaf)
        self.fill_api_version(cmd_portal_info, target_version)
        self.generate_base_learn_more(cmd_portal_info, leaf)
        cmd_info = {}
        self.generate_cmd_info(cmd_info, cmd_cfg, leaf, target_version)
        cmd_portal_info['commands'] = [cmd_info]
        return cmd_portal_info

    def generate_command_portal_raw(self, cmd_cfg, leaf, target_version):
        if not isinstance(cmd_cfg, CMDCommand) or \
                not isinstance(leaf, CMDSpecsCommand) or \
                not isinstance(target_version, CMDSpecsCommandVersion):
            return None
        if cmd_cfg.arg_groups is None:
            logging.info("{0} cmd has no arg_group".format(" ".join(leaf.names)))
            return None
        if not self.__check_cmd_examples(cmd_cfg):
            logging.info("{0} cmd do not has examples".format(" ".join(leaf.names)))
            return None
        cmd_portal_info = {}
        self.fill_resource_type(cmd_portal_info, cmd_cfg, leaf)
        self.fill_api_version(cmd_portal_info, target_version)
        self.generate_base_learn_more(cmd_portal_info, leaf)
        self.generate_cmd_info(cmd_portal_info, cmd_cfg, leaf, target_version)
        return cmd_portal_info

    def get_portal_file_path(self):
        cli_folder = Config.CLI_PATH
        if not os.path.exists(cli_folder) or not os.path.isdir(cli_folder):
            raise ValueError(f"Invalid Cli Main Repo folder: '{cli_folder}'")
        data_model_folder = os.path.join(cli_folder, "data-model-for-portal")
        return data_model_folder

    def valid_portal_json_format(self, portal_data):
        try:
            jsonschema.validate(instance=portal_data, schema=portal_file_schema.PORTAL_FILE_SCHEMA)
        except jsonschema.exceptions.ValidationError as err:
            #print(err)
            return False
        return True

    def generate_portal_file(self, portal_dict):
        data_model_folder = self.get_portal_file_path()
        for rp_name, resource_infos in portal_dict.items():
            for resource_type, resource_info in resource_infos.items():
                resource_paths = [rp_name] + resource_type.split("/")
                resource_file_path = data_model_folder + "/" + "/".join(resource_paths[:-1]) + "/" + \
                                     resource_paths[-1] + ".json"
                resource_file_folder = os.path.dirname(resource_file_path)
                if not os.path.exists(resource_file_folder):
                    os.makedirs(resource_file_folder)
                if not self.valid_portal_json_format(resource_info):
                    logging.info("json format error")
                    continue
                with open(resource_file_path, "w") as f_out:
                    f_out.write(json.dumps(resource_info, indent=4))

    def generate_cmds_portal_file(self, cmd_portal_list):
        portal_dict = {}
        cmd_dedup = set()
        for cmd_portal_info in cmd_portal_list:
            if 'rp_name' not in cmd_portal_info or 'resourceType' not in cmd_portal_info or \
                'apiVersion' not in cmd_portal_info or 'name' not in cmd_portal_info:
                continue

            cmd_name = cmd_portal_info['name']
            if cmd_name in cmd_dedup:
                logging.info("{0} cmd has repeated from versions {1}".format(cmd_name, cmd_portal_info['apiVersion']))
                continue
            cmd_dedup.add(cmd_name)
            rp_name = cmd_portal_info['rp_name']
            rescoure_type = cmd_portal_info['resourceType']
            api_version = cmd_portal_info['apiVersion']
            learn_more = cmd_portal_info['learnMore']

            del cmd_portal_info['rp_name'], cmd_portal_info['resourceType']
            del cmd_portal_info['apiVersion'], cmd_portal_info['learnMore']
            if rp_name not in portal_dict:
                portal_dict[rp_name] = {
                    rescoure_type: {
                        "resourceType": rescoure_type,
                        "apiVersion": api_version,
                        "learnMore": learn_more,
                        'commands': [cmd_portal_info]
                    }
                }
                continue
            rescoure_infos = portal_dict[rp_name]
            if rescoure_type not in rescoure_infos:
                rescoure_infos[rescoure_type] = {
                    "resourceType": rescoure_type,
                    "apiVersion": api_version,
                    "learnMore": learn_more,
                    'commands': [cmd_portal_info]
                }
                continue
            rescoure_info = rescoure_infos[rescoure_type]
            rescoure_info['commands'].append(cmd_portal_info)

        self.generate_portal_file(portal_dict)

    def generate_cmds_portal_info(self, aaz_spec_manager, registered_cmds):
        cmd_portal_list = []
        for cmd_name_version in registered_cmds:
            # cmd_name_version = ['monitor', 'diagnostic-setting', 'list', '2021-05-01-preview']
            node_names = cmd_name_version[:-2]
            leaf_name = cmd_name_version[-2]
            registered_version = cmd_name_version[-1]
            leaf = aaz_spec_manager.find_command(*node_names, leaf_name)
            if not leaf or not leaf.versions:
                logging.warning("Command group: " + " ".join(node_names) + " not exist")
                continue
            if not leaf.versions:
                logging.warning("Command group: " + " ".join(node_names) + " version not exist")
                continue
            target_version = None
            for v in (leaf.versions or []):
                if v.name == registered_version:
                    target_version = v
                    break
            if not target_version:
                logging.warning("Command: " + " ".join(node_names) + " version not exist")
                continue
            logging.info("Generating portal config of [ az {0} ] with registered version {1}".format(" ".join(cmd_name_version[:-1]),
                                                                                              registered_version))
            cfg_reader = aaz_spec_manager.load_resource_cfg_reader_by_command_with_version(leaf, version=target_version)
            cmd_cfg = cfg_reader.find_command(*leaf.names)
            cmd_portal_info = self.generate_command_portal_raw(cmd_cfg, leaf, target_version)
            if cmd_portal_info:
                cmd_portal_list.append(cmd_portal_info)
        return cmd_portal_list

if __name__ == '__main__':
    generator = PortalCliGenerator()
    resource_info = {
        "resourceType": "virtualMachines",
        "apiVersion": "2021-11-01",
        "confirmation": True,
        "learnMore": {
            "url": "https://docs.microsoft.com/cli/azure/vm"
        },
    }
    schema = {
        "type": "object",
        "properties": {
            "resourceType": {"type": "string"},
            "apiVersion": {"type": "string"},
            "confirmation": {"type": "boolean"},
            "learnMore": {
                "type": "object",
                "properties": {
                    "url": {"type": "string"}
                }
            }
        }
    }
    print(jsonschema.validate(instance=resource_info, schema=schema))
    try:
        jsonschema.validate(instance=resource_info, schema=schema)
    except jsonschema.exceptions.ValidationError as err:
        print(err)

    print("verification done")





from schematics.types import StringType
from utils.plane import PlaneEnum


class PlaneField(StringType):

    def __init__(self, *args, **kwargs):
        super(PlaneField, self).__init__(
            choices=PlaneEnum.choices(),
            *args, **kwargs
        )


__all__ = ["PlaneField"]



from flask import Blueprint, jsonify, request, url_for
from utils import exceptions
from command.controller.specs_manager import AAZSpecsManager


bp = Blueprint('specs', __name__, url_prefix='/AAZ/Specs')


# modules
@bp.route("/CommandTree/Nodes/<names_path:node_names>", methods=("GET",))
def command_tree_node(node_names):
    if node_names[0] != AAZSpecsManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command group not exist")
    node_names = node_names[1:]

    manager = AAZSpecsManager()
    node = manager.find_command_group(*node_names)
    if not node:
        raise exceptions.ResourceNotFind("Command group not exist")

    result = node.to_primitive()
    return jsonify(result)


@bp.route("/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>", methods=("GET",))
def command_tree_leaf(node_names, leaf_name):
    if node_names[0] != AAZSpecsManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = AAZSpecsManager()
    leaf = manager.find_command(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")

    result = leaf.to_primitive()
    return jsonify(result)


@bp.route("/CommandTree/Nodes/<names_path:node_names>/Leaves/<name:leaf_name>/Versions/<base64:version_name>", methods=("GET",))
def aaz_command_in_version(node_names, leaf_name, version_name):
    if node_names[0] != AAZSpecsManager.COMMAND_TREE_ROOT_NAME:
        raise exceptions.ResourceNotFind("Command not exist")
    node_names = node_names[1:]

    manager = AAZSpecsManager()
    leaf = manager.find_command(*node_names, leaf_name)
    if not leaf:
        raise exceptions.ResourceNotFind("Command not exist")

    version = None
    for v in (leaf.versions or []):
        if v.name == version_name:
            version = v
            break

    if not version:
        raise exceptions.ResourceNotFind("Command of version not exist")

    cfg_reader = manager.load_resource_cfg_reader_by_command_with_version(leaf, version=version)
    cmd_cfg = cfg_reader.find_command(*leaf.names)

    result = cmd_cfg.to_primitive()
    del result['name']
    result.update({
        'names': leaf.names,
        'help': leaf.help.to_primitive(),
        'stage': version.stage,
    })
    if version.examples:
        result['examples'] = version.examples.to_primitive()

    return jsonify(result)


@bp.route("/Resources/<plane>/<base64:resource_id>", methods=("GET", ))
def get_resource(plane, resource_id):
    manager = AAZSpecsManager()
    versions = manager.get_resource_versions(plane, resource_id)
    if versions is None:
        raise exceptions.ResourceNotFind("Resource not exist")
    result = {
        "id": resource_id,
        "versions": versions
    }
    return jsonify(result)


@bp.route("/Resources/<plane>/Filter", methods=("Post", ))
def filter_resources(plane):
    data = request.get_json()
    if 'resources' not in data:
        raise exceptions.InvalidAPIUsage("Invalid request body")
    manager = AAZSpecsManager()

    result = {
        'resources': []
    }
    for resource_id in data['resources']:
        versions = manager.get_resource_versions(plane, resource_id)
        if versions is None:
            continue
        result['resources'].append({
            "id": resource_id,
            "versions": versions,
        })

    return jsonify(result)



from command.model.configuration._fields import CMDResourceIdField, CMDVersionField
from schematics.models import Model
from schematics.types import StringType
from utils.fields import PlaneField


class CMDSpecsResource(Model):
    plane = PlaneField(required=True)
    id = CMDResourceIdField(required=True)
    version = CMDVersionField(required=True)
    subresource = StringType()

    class Options:
        serialize_when_none = False



from schematics.models import Model
from schematics.types import StringType, BooleanType, BaseType, ModelType, ListType


class XmsEnumValue(Model):
    """
    When set, this will override the values specified with enum, while also enabling further customization. We recommend still specifying enum as a fallback for consumers that don't understand x-ms-enum. Each item in x-ms-enum corresponds to an enum item.
    """

    value = BaseType(required=True)  # Property value is mandatory and corresponds to the value one would also have specified using enum.
    description = StringType()  # Property value is mandatory and corresponds to the value one would also have specified using enum
    name = StringType()  # allows overriding the name of the enum value that would usually be derived from the value.


class XmsEnum(Model):
    """
    Enum definitions in OpenAPI indicate that only a particular set of values may be used for a property or parameter. When the property is represented on the wire as a string, it would be a natural choice to represent the property type in C# and Java as an enum. However, not all enumeration values should necessarily be represented as strongly typed enums - there are additional considerations, such as how often expected values might change, since adding a new value to a strongly typed enum is a breaking change requiring an updated API version. Additionally, there is some metadata that is required to create a useful enum, such as a descriptive name, which is not represented in vanilla OpenAPI. For this reason, enums are not automatically turned into strongly typed enum types - instead they are rendered in the documentation comments for the property or parameter to indicate allowed values. To indicate that an enum will rarely change and that C#/Java enum semantics are desired, use the x-ms-enum extension. Note that depending on the code generation language the behavior of this extension may differ.
    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-enum
    """

    name = StringType(required=True)  # Specifies the name for the Enum.
    model_as_string = BooleanType(
        default=False,
        serialized_name="modelAsString",
        deserialize_from="modelAsString"
    )  # When set to true the enum will be modeled as a string. No validation will happen. When set to false, it will be modeled as an enum if that language supports enums. Validation will happen, irrespective of support of enums in that language.
    values = ListType(ModelType(XmsEnumValue))

    model_as_extensible = BooleanType(
        serialized_name="modelAsExtensible",
        deserialize_from="modelAsExtensible"
    )   # TODO: don't know its usage


class XmsEnumField(ModelType):

    def __init__(self, **kwargs):
        super(XmsEnumField, self).__init__(
            XmsEnum,
            serialized_name="x-ms-enum",
            deserialize_from="x-ms-enum",
            **kwargs
        )



from werkzeug.routing import BaseConverter, PathConverter

from utils.base64 import b64decode_str, b64encode_str


class Base64Converter(BaseConverter):

    def to_python(self, value: str) -> str:
        return b64decode_str(value)

    def to_url(self, value: str) -> str:
        return super(Base64Converter, self).to_url(b64encode_str(value))


class NameConverter(BaseConverter):
    regex = r"[a-z0-9]+(-[a-z0-9]+)*"


class NameWithCapitalConverter(BaseConverter):
    regex = r"[a-zA-Z0-9]+(-[a-zA-Z0-9]+)*"


class NamesPathConverter(BaseConverter):
    regex = r"[a-z0-9]+(-[a-z0-9]+)*(/[a-z0-9]+(-[a-z0-9]+)*)*"
    weight = 200

    def to_python(self, value):
        return value.split('/')

    def to_url(self, values):
        if isinstance(values, str):
            return super(NamesPathConverter, self).to_url(values)
        return '/'.join(super(NamesPathConverter, self).to_url(value) for value in values)


class ListPathConvertor(PathConverter):

    def to_python(self, value):
        return value.split('/')

    def to_url(self, values):
        if isinstance(values, str):
            return super(ListPathConvertor, self).to_url(values)
        return '/'.join(super(ListPathConvertor, self).to_url(value) for value in values)


__all__ = ["Base64Converter", "NameConverter", "NamesPathConverter", "ListPathConvertor"]



from schematics.models import Model
from schematics.types import PolyModelType, ListType
from schematics.types.serializable import serializable

from ._fields import CMDVariantField


class CMDConditionOperator(Model):
    # properties as tags
    TYPE_VALUE = None

    @serializable
    def type(self):
        return self._get_type()

    def _get_type(self):
        assert self.TYPE_VALUE is not None
        return self.TYPE_VALUE

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.TYPE_VALUE is None:
            return False

        if isinstance(data, dict):
            type_value = data.get('type', None)
            return type_value == cls.TYPE_VALUE
        elif isinstance(data, CMDConditionOperator):
            return data.TYPE_VALUE == cls.TYPE_VALUE
        return False


class CMDConditionAndOperator(CMDConditionOperator):
    TYPE_VALUE = "and"

    # properties as nodes
    operators = ListType(PolyModelType(CMDConditionOperator, allow_subclasses=True), min_size=2)


class CMDConditionOrOperator(CMDConditionOperator):
    TYPE_VALUE = "or"

    operators = ListType(PolyModelType(CMDConditionOperator, allow_subclasses=True), min_size=2)


class CMDConditionNotOperator(CMDConditionOperator):
    TYPE_VALUE = "not"

    # properties as nodes
    operator = PolyModelType(CMDConditionOperator, allow_subclasses=True, required=True)


class CMDConditionHasValueOperator(CMDConditionOperator):
    TYPE_VALUE = "hasValue"

    # properties as tags
    arg = CMDVariantField(required=True)


class CMDCondition(Model):
    # properties as tags
    var = CMDVariantField(required=True)  # define variant

    # properties as nodes
    operator = PolyModelType(CMDConditionOperator, allow_subclasses=True, required=True)



class CMDDiffLevelEnum:

    BreakingChange = 1  # diff breaking change part
    Structure = 2    #
    Associate = 5    # include diff for links
    All = 10    # including description and help messages


class CMDArgBuildPrefix:
    Query = '$Query'
    Header = '$Header'
    Path = '$Path'


DEFAULT_CONFIRMATION_PROMPT = "Are you sure you want to perform this operation?"



import logging

import inflect
from command.model.configuration import CMDCommandGroup, CMDCommand, CMDHttpOperation, CMDHttpRequest, \
    CMDSchemaDefault, CMDHttpResponseJsonBody, CMDArrayOutput, CMDJsonInstanceUpdateAction, \
    CMDInstanceUpdateOperation, CMDRequestJson, DEFAULT_CONFIRMATION_PROMPT, CMDClsSchemaBase, CMDHttpResponse, \
    CMDResponseJson
from swagger.model.schema.cmd_builder import CMDBuilder
from swagger.model.schema.fields import MutabilityEnum
from swagger.model.schema.path_item import PathItem
from swagger.model.specs import SwaggerLoader
from swagger.model.specs._utils import operation_id_separate, camel_case_to_snake_case, get_url_path_valid_parts
from utils import exceptions
from utils.config import Config
from utils.plane import PlaneEnum
from utils.error_format import AAZErrorFormatEnum

logger = logging.getLogger('backend')


class BuildInVariants:

    Instance = "$Instance"


class CommandGenerator:
    _inflect_engine = inflect.engine()

    def __init__(self):
        self.loader = SwaggerLoader()

    def load_resources(self, resources):
        for resource in resources:
            self.loader.load_file(resource.file_path)
        self.loader.link_swaggers()

    def create_draft_command_group(self, resource,
                                   update_by=None,
                                   methods=('get', 'delete', 'put', 'post', 'head', 'patch'),
                                   **kwargs):
        swagger = self.loader.get_loaded(resource.file_path)
        assert swagger is not None
        path_item = swagger.paths.get(resource.path, None)
        if path_item is None:
            path_item = swagger.x_ms_paths.get(resource.path, None)

        command_group = CMDCommandGroup()
        command_group.commands = []

        assert isinstance(path_item, PathItem)
        if path_item.get is not None and 'get' in methods:
            cmd_builder = CMDBuilder(path=resource.path, method='get', mutability=MutabilityEnum.Read)
            show_or_list_command = self.generate_command(path_item, resource, cmd_builder)
            command_group.commands.append(show_or_list_command)

        if path_item.delete is not None and 'delete' in methods:
            cmd_builder = CMDBuilder(path=resource.path, method='delete', mutability=MutabilityEnum.Create)
            delete_command = self.generate_command(path_item, resource, cmd_builder)
            delete_command.confirmation = DEFAULT_CONFIRMATION_PROMPT   # add confirmation for delete command by default
            command_group.commands.append(delete_command)

        if path_item.put is not None and 'put' in methods:
            cmd_builder = CMDBuilder(path=resource.path, method='put', mutability=MutabilityEnum.Create)
            create_command = self.generate_command(path_item, resource, cmd_builder)
            command_group.commands.append(create_command)

        if path_item.post is not None and 'post' in methods:
            cmd_builder = CMDBuilder(path=resource.path, method='post', mutability=MutabilityEnum.Create)
            action_command = self.generate_command(path_item, resource, cmd_builder)
            command_group.commands.append(action_command)

        if path_item.head is not None and 'head' in methods:
            cmd_builder = CMDBuilder(path=resource.path, method='head', mutability=MutabilityEnum.Read)
            head_command = self.generate_command(path_item, resource, cmd_builder)
            command_group.commands.append(head_command)

        # update command
        if update_by is None:
            update_by_patch_command = None
            update_by_generic_command = None
            if path_item.patch is not None and 'patch' in methods:
                cmd_builder = CMDBuilder(path=resource.path, method='patch', mutability=MutabilityEnum.Update)
                update_by_patch_command = self.generate_command(path_item, resource, cmd_builder)
            if path_item.get is not None and path_item.put is not None and 'get' in methods and 'put' in methods:
                cmd_builder = CMDBuilder(path=resource.path)
                update_by_generic_command = self.generate_generic_update_command(path_item, resource, cmd_builder)
            # generic update command first, patch update command after that
            if update_by_generic_command:
                command_group.commands.append(update_by_generic_command)
            elif update_by_patch_command:
                command_group.commands.append(update_by_patch_command)
        else:
            if update_by == 'GenericOnly':
                if path_item.get is None or path_item.put is None:
                    raise exceptions.InvalidAPIUsage(f"Invalid update_by resource: resource needs to have 'get' and 'put' operations: '{resource}'")
                if 'get' not in methods or 'put' not in methods:
                    raise exceptions.InvalidAPIUsage(f"Invalid update_by resource: '{resource}': 'get' or 'put' not in methods: '{methods}'")
                cmd_builder = CMDBuilder(path=resource.path)
                generic_update_command = self.generate_generic_update_command(path_item, resource, cmd_builder)
                if generic_update_command is None:
                    raise exceptions.InvalidAPIUsage(f"Invalid update_by resource: failed to generate generic update: '{resource}'")
                command_group.commands.append(generic_update_command)
                # elif 'update_by' in kwargs:
                #     logger.error(f'Failed to generate generic update for resource: {resource}')
            elif update_by == 'PatchOnly':
                if path_item.patch is None:
                    raise exceptions.InvalidAPIUsage(f"Invalid update_by resource: resource needs to have 'patch' operation: '{resource}'")
                if 'patch' not in methods:
                    raise exceptions.InvalidAPIUsage(f"Invalid update_by resource: '{resource}': 'patch' not in methods: '{methods}'")
                cmd_builder = CMDBuilder(path=resource.path, method='patch', mutability=MutabilityEnum.Update)
                patch_update_command = self.generate_command(path_item, resource, cmd_builder)
                command_group.commands.append(patch_update_command)
            # elif update_by == 'GenericAndPatch':
            #     # TODO: add support for generic and patch merge
            #     if path_item.get is None or path_item.put is None or path_item.patch is None:
            #         raise exceptions.InvalidAPIUsage(f"Invalid update_by resource: resource needs to have 'get' and 'put' and 'patch' operation: '{resource}'")
            #     if 'get' not in methods or 'put' not in methods or 'patch' not in methods:
            #         raise exceptions.InvalidAPIUsage(f"Invalid update_by resource: '{resource}': 'get' or 'put' or 'patch' not in methods: '{methods}'")
            #     cmd_builder = CMDBuilder(path=resource.path)
            #     generic_update_command = self.generate_generic_update_command(path_item, resource, cmd_builder)
            #     if generic_update_command is None:
            #         raise exceptions.InvalidAPIUsage(f"Invalid update_by resource: failed to generate generic update: '{resource}'")
            #     cmd_builder = CMDBuilder(path=resource.path, method='patch', mutability=MutabilityEnum.Update)
            #     patch_update_command = self.generate_command(path_item, resource, cmd_builder)
            #     generic_and_patch_update_command = self._merge_update_commands(
            #         patch_command=patch_update_command, generic_command=generic_update_command
            #     )
            #     command_group.commands.append(generic_and_patch_update_command)
            elif update_by != 'None':
                raise exceptions.InvalidAPIUsage(f"Invalid update_by value: {update_by} : only support ['GenericOnly', 'PatchOnly', 'None'] values")

        for command in command_group.commands:
            parts = command.name.split(' ')
            group_name = ' '.join(parts[:-1])
            if command_group.name:
                assert group_name == command_group.name
            else:
                command_group.name = group_name
            command.name = parts[-1]  # remove the command group name parts
            self.optimize_command_description(command)

        return command_group

    @staticmethod
    def generate_command_version(resource):
        return resource.version

    def generate_command(self, path_item, resource, cmd_builder):
        command = CMDCommand()
        command.version = self.generate_command_version(resource)
        command.resources = [
            resource.to_cmd()
        ]

        op = self._generate_operation(cmd_builder, path_item)
        cmd_builder.apply_cls_definitions(op)

        assert isinstance(op, CMDHttpOperation)
        if not self._set_api_version_parameter(op.http.request, api_version=resource.version):
            logger.warning(
                f"Cannot Find api version parameter: {cmd_builder.path}, '{cmd_builder.method}' : {path_item.traces}")

        command.description = op.description
        command.operations = [op]

        command.generate_args()
        command.generate_outputs(pageable=cmd_builder.get_pageable(path_item, op))

        output = command.outputs[0] if command.outputs else None
        command.name = self._generate_command_name(path_item, resource, cmd_builder.method, output)

        return command

    def generate_generic_update_command(self, path_item, resource, cmd_builder):
        command = CMDCommand()
        command.version = self.generate_command_version(resource)
        command.resources = [
            resource.to_cmd()
        ]
        assert path_item.get is not None
        assert path_item.put is not None

        get_op = self._generate_operation(cmd_builder, path_item, method='get', mutability=MutabilityEnum.Read)
        put_op = self._generate_operation(cmd_builder, path_item, method='put', mutability=MutabilityEnum.Update)

        cmd_builder.apply_cls_definitions(get_op, put_op)

        if put_op.http.request.body is None:
            return None

        if not self._set_api_version_parameter(get_op.http.request, api_version=resource.version):
            logger.warning(f"Cannot Find api version parameter: {resource.path}, 'get' : {path_item.traces}")
        if not self._set_api_version_parameter(put_op.http.request, api_version=resource.version):
            logger.warning(f"Cannot Find api version parameter: {resource.path}, 'put' : {path_item.traces}")

        if not command.build_output_by_operation(get_op):
            return None

        if not command.build_output_by_operation(put_op):
            return None

        self._filter_generic_update_parameters(get_op, put_op)

        command.description = put_op.description
        json_update_op = self._generate_instance_update_operation(put_op)
        command.operations = [
            get_op,
            json_update_op,
            put_op
        ]

        command.generate_args()
        command.generate_outputs()

        assert command.outputs

        group_name = self.generate_command_group_name_by_resource(
            resource_path=resource.path, rp_name=resource.resource_provider.name)
        command.name = f"{group_name} update"
        return command

    @staticmethod
    def _generate_operation(cmd_builder, path_item, **kwargs):
        op = cmd_builder(path_item, **kwargs)

        assert isinstance(op, CMDHttpOperation)
        error_format = None
        for resp in op.http.responses:
            if resp.is_error:
                if not isinstance(resp.body, CMDHttpResponseJsonBody):
                    if not resp.body:
                        raise exceptions.InvalidAPIUsage(
                            f"Invalid `Error` response schema in operation `{op.operation_id}`: "
                            f"Missing `schema` property in response "
                            f"`{resp.status_codes or 'default'}`."
                        )
                    else:
                        raise exceptions.InvalidAPIUsage(
                            f"Invalid `Error` response schema in operation `{op.operation_id}`: "
                            f"Only support json schema, current is '{type(resp.body)}' in response "
                            f"`{resp.status_codes or 'default'}`"
                        )
                schema = resp.body.json.schema
                if not isinstance(schema, CMDClsSchemaBase):
                    raise NotImplementedError()
                name = schema.type[1:]
                if not error_format:
                    error_format = name
                if error_format != name:
                    raise exceptions.InvalidAPIUsage(
                        f"Invalid `Error` response schema in operation `{op.operation_id}`: "
                        f"Multiple schema formats are founded: {name}, {error_format}"
                    )
            else:
                if resp.body is None:
                    continue
                if isinstance(resp.body, CMDHttpResponseJsonBody):
                    resp.body.json.var = BuildInVariants.Instance

        if not error_format:
            if Config.DEFAULT_PLANE != PlaneEnum.Mgmt:
                raise exceptions.InvalidAPIUsage(
                    f"Missing `Error` response schema in operation `{op.operation_id}`: "
                    f"Please define the `default` response in swagger for error."
                )
            # use MgmtErrorFormat for default error response schema
            error_format = AAZErrorFormatEnum.MgmtErrorFormat
            err_response = CMDHttpResponse()
            err_response.is_error = True
            err_response.status_codes = []
            err_response.body = CMDHttpResponseJsonBody()
            err_response.body.json = CMDResponseJson()
            err_schema = CMDClsSchemaBase()
            err_schema._type = f"@{error_format}"
            err_response.body.json.schema = err_schema
            op.http.responses.append(err_response)
        elif not AAZErrorFormatEnum.validate(error_format):
            raise exceptions.InvalidAPIUsage(
                f"Invalid `Error` response schema in operation `{op.operation_id}`: "
                f"Invalid error format `{error_format}`. Support `ODataV4Format` and `MgmtErrorFormat` only"
            )

        return op

    @staticmethod
    def _set_api_version_parameter(request, api_version):
        assert isinstance(request, CMDHttpRequest)
        assert isinstance(api_version, str)
        find_api_version = False
        query = request.query
        if query is not None:
            for idx in range(len(query.params)):
                param = query.params[idx]
                if param.name == "api-version":
                    param.default = CMDSchemaDefault()
                    param.default.value = api_version
                    param.read_only = True
                    param.const = True
                    if query.consts is None:
                        query.consts = []
                    query.consts.append(param)
                    query.params.pop(idx)
                    find_api_version = True
                    break
        return find_api_version

    @classmethod
    def generate_command_group_name_by_resource(cls, resource_path, rp_name):
        valid_parts = get_url_path_valid_parts(resource_path, rp_name)

        names = []

        # add resource provider name as command group name
        for rp_part in rp_name.split('.'):
            if rp_part.lower() in ("microsoft", "azure"):
                # ignore microsoft and azure keywards
                continue
            names.append(camel_case_to_snake_case(rp_part, '-'))

        for part in valid_parts[1:]:  # ignore first part to avoid include resource provider
            if part.startswith('{'):
                continue
            name = camel_case_to_snake_case(part, '-')
            singular_name = cls._inflect_engine.singular_noun(name) or name
            names.append(singular_name)
        return " ".join([name for name in names if name])

    def _generate_command_name(self, path_item, resource, method, output):
        group_name = self.generate_command_group_name_by_resource(
            resource_path=resource.path, rp_name=resource.resource_provider.name)
        url_path = resource.id.split("?")[0]
        if method == "get":
            if url_path.endswith("/{}"):
                command_name = f"{group_name} show"
            else:
                sub_url_path = url_path + "/{}"
                if path_item.get.x_ms_pageable:
                    command_name = f"{group_name} list"
                elif isinstance(output, CMDArrayOutput) and output.next_link is not None:
                    command_name = f"{group_name} list"
                    # logger.debug(
                    #     f"Command Name For Get set to 'list' by nexLink: {resource.path} :"
                    #     f" {path_item.get.operation_id} : {path_item.traces}"
                    # )
                elif sub_url_path in resource.resource_provider.get_resource_map():
                    command_name = f"{group_name} list"
                    # logger.debug(
                    #     f"Command Name For Get set to 'list' by sub_url_path: {resource.path} :"
                    #     f" {path_item.get.operation_id} : {path_item.traces}"
                    # )
                else:
                    # by operation id
                    op_parts = operation_id_separate(path_item.get.operation_id)
                    contain_list = False
                    contain_get = False
                    for part in op_parts:
                        if "list" in part:
                            contain_list = True
                        elif "get" in part:
                            contain_get = True
                    if contain_list and not contain_get:
                        command_name = f"{group_name} list"
                        # logger.debug(
                        #     f"Command Name For Get set to 'list' by operation_id: {resource.path} :"
                        #     f" {path_item.get.operation_id} : {path_item.traces}"
                        # )
                    elif contain_get and not contain_list:
                        command_name = f"{group_name} show"
                        # logger.debug(
                        #     f"Command Name For Get set to 'show' by operation_id: {resource.path} :"
                        #     f" {path_item.get.operation_id} : {path_item.traces}"
                        # )
                    else:
                        command_name = f"{group_name} " + '-'.join(op_parts[-1])
                        logger.warning(
                            f"Command Name For Get set by operation_id: {command_name} : {resource.path} :"
                            f" {path_item.get.operation_id} : {path_item.traces}"
                        )

        elif method == "delete":
            command_name = f"{group_name} delete"
        elif method == "put":
            command_name = f"{group_name} create"
        elif method == "patch":
            command_name = f"{group_name} update"
        elif method == "head":
            command_name = f"{group_name} head"
        elif method == "post":
            if not url_path.endswith("/{}") and not path_item.get and not path_item.put and not path_item.patch and \
                    not path_item.delete and not path_item.head:
                # directly use group name as command name
                command_name = group_name
            else:
                op_parts = operation_id_separate(path_item.post.operation_id)
                command_name = f"{group_name} " + '-'.join(op_parts[-1])
                logger.warning(f"Command Name For Post set by operation_id: {command_name} : {resource.path} :"
                               f" {path_item.post.operation_id} : {path_item.traces}")
        else:
            raise NotImplementedError()
        return command_name

    # For update
    @staticmethod
    def _generate_instance_update_operation(put_op):
        json_update_op = CMDInstanceUpdateOperation()
        json_update_op.instance_update = CMDJsonInstanceUpdateAction()
        json_update_op.instance_update.ref = BuildInVariants.Instance
        json_update_op.instance_update.json = CMDRequestJson()
        json_update_op.instance_update.json.schema = put_op.http.request.body.json.schema

        put_op.http.request.body.json.ref = BuildInVariants.Instance
        put_op.http.request.body.json.schema = None
        return json_update_op

    @staticmethod
    def _filter_generic_update_parameters(get_op, put_op):
        """Get operation may contains useless query or header parameters for update, ignore them"""
        get_request = get_op.http.request
        put_request = put_op.http.request

        query_name_set = set()
        if put_request.query:
            for param in put_request.query.params:
                query_name_set.add(param.name)
        if get_request.query:
            get_query_params = []
            for param in get_request.query.params:
                if param.name in query_name_set:
                    get_query_params.append(param)
                elif param.required:
                    print(f"Query param {param.name} in Get ({get_op.operation_id}) not in Put ({put_op.operation_id})")
                    get_query_params.append(param)
            get_request.query.params = get_query_params

        header_name_set = set()
        if put_request.header:
            for param in put_request.header.params:
                header_name_set.add(param.name)
        if get_request.header:
            get_header_params = []
            for param in get_request.header.params:
                if param.name in header_name_set:
                    get_header_params.append(param)
                elif param.required:
                    print(
                        f"Header param {param.name} in Get ({get_op.operation_id}) not in Put ({put_op.operation_id})")
                    get_header_params.append(param)
            get_request.header.params = get_header_params

    @staticmethod
    def _merge_update_commands(patch_command, generic_command):
        # TODO: merge patch command and generic command into one
        return generic_command

    @staticmethod
    def optimize_command_description(command):
        if command.description:
            keywords = command.description.split(' ')
            if command.name.lower() == "show":
                keywords[0] = "Get"
            elif command.name.lower() == "list":
                keywords[0] = "List"
            elif command.name.lower() == "delete":
                keywords[0] = "Delete"
            elif command.name.lower() == "create":
                if len(keywords) > 1 and keywords[1].lower() == "or":
                    keywords = ["Create", *keywords[3:]]
                else:
                    keywords[0] = "Create"
            elif command.name.lower() == "update":
                if len(keywords) > 1 and keywords[1].lower() == "or":
                    keywords = ["Update", *keywords[3:]]
                else:
                    keywords[0] = "Update"
            command.description = " ".join(keywords)





class AAZErrorFormatEnum:

    ODataV4Format = "ODataV4Format"  # registered in azure.cli.core.aaz._error_format
    MgmtErrorFormat = "MgmtErrorFormat"  # registered in azure.cli.core.aaz._error_format

    @classmethod
    def validate(cls, format):
        return format in (cls.ODataV4Format, cls.MgmtErrorFormat)

    @classmethod
    def classify_error_format(cls, builder, schema):
        from command.model.configuration import CMDObjectSchemaBase, CMDObjectSchema, CMDClsSchemaBase
        if isinstance(schema, CMDClsSchemaBase):
            schema = builder.get_cls_definition_model(schema)
        if not isinstance(schema, CMDObjectSchemaBase) or not schema.props:
            return None

        props = {}
        for prop in schema.props:
            props[prop.name] = prop

        if "error" in props:
            schema = props['error']
            if isinstance(schema, CMDClsSchemaBase):
                schema = builder.get_cls_definition_model(schema)
            if not isinstance(schema, CMDObjectSchemaBase):
                return None

        prop_keys = set()
        for prop in schema.props:
            prop_keys.add(prop.name)

        if not ("code" in prop_keys and "message" in prop_keys):
            # code and message is required
            return None

        prop_keys.difference_update({"code", "message", "target", "details", "innerError", "innererror"})  # some api use "innerError" instead of "innererror"
        if len(prop_keys) == 0:
            return cls.ODataV4Format
        if prop_keys == {"additionalInfo"}:
            return cls.MgmtErrorFormat

        return None



import json
import logging
import os
import shutil
from datetime import datetime

from command.model.editor import CMDEditorWorkspace, CMDCommandTreeNode, CMDCommandTreeLeaf
from swagger.controller.command_generator import CommandGenerator
from swagger.controller.specs_manager import SwaggerSpecsManager
from swagger.utils.exceptions import InvalidSwaggerValueError
from utils import exceptions
from utils.config import Config
from .specs_manager import AAZSpecsManager
from .workspace_cfg_editor import WorkspaceCfgEditor
from command.model.configuration import CMDHelp, CMDResource, CMDCommandExample, CMDArg, CMDCommand

logger = logging.getLogger('backend')


class WorkspaceManager:
    COMMAND_TREE_ROOT_NAME = "aaz"

    IN_MEMORY = "__IN_MEMORY_WORKSPACE__"

    @classmethod
    def list_workspaces(cls):
        workspaces = []
        if not os.path.exists(Config.AAZ_DEV_WORKSPACE_FOLDER):
            return workspaces

        for name in os.listdir(Config.AAZ_DEV_WORKSPACE_FOLDER):
            if not os.path.isdir(os.path.join(Config.AAZ_DEV_WORKSPACE_FOLDER, name)):
                continue
            manager = cls(name)
            if os.path.exists(manager.path) and os.path.isfile(manager.path):
                workspaces.append({
                    "name": name,
                    "folder": manager.folder,
                    "updated": os.path.getmtime(manager.path)
                })
        return workspaces

    @classmethod
    def new(cls, name, plane, **kwargs):
        manager = cls(name, **kwargs)
        if not manager.is_in_memory and os.path.exists(manager.path):
            raise exceptions.ResourceConflict(f"Workspace conflict: Workspace json file path exists: {manager.path}")
        manager.ws = CMDEditorWorkspace({
            "name": name,
            "plane": plane,
            "version": datetime.utcnow(),
            "commandTree": {
                "names": [cls.COMMAND_TREE_ROOT_NAME],
            }
        })
        return manager

    def __init__(self, name, folder=None, aaz_manager=None, swagger_manager=None):
        self.name = name
        if not folder:
            if not Config.AAZ_DEV_WORKSPACE_FOLDER or os.path.exists(Config.AAZ_DEV_WORKSPACE_FOLDER) and not os.path.isdir(Config.AAZ_DEV_WORKSPACE_FOLDER):
                raise ValueError(
                    f"Invalid AAZ_DEV_WORKSPACE_FOLDER: Expect a folder path: {Config.AAZ_DEV_WORKSPACE_FOLDER}")
            self.folder = os.path.join(Config.AAZ_DEV_WORKSPACE_FOLDER, name)
        else:
            self.folder = os.path.expanduser(folder) if folder != self.IN_MEMORY else self.IN_MEMORY
        if not self.is_in_memory and os.path.exists(self.folder) and not os.path.isdir(self.folder):
            raise ValueError(f"Invalid workspace folder: Expect a folder path: {self.folder}")
        self.path = os.path.join(self.folder, 'ws.json')

        self.ws = None
        self._cfg_editors = {}
        self._reusable_leaves = {}

        self.aaz_specs = aaz_manager or AAZSpecsManager()
        self.swagger_specs = swagger_manager or SwaggerSpecsManager()
        self.swagger_command_generator = CommandGenerator()

    @property
    def is_in_memory(self):
        return self.folder == self.IN_MEMORY

    def load(self):
        assert not self.is_in_memory
        # TODO: handle exception
        if not os.path.exists(self.path) or not os.path.isfile(self.path):
            raise exceptions.ResourceNotFind(f"Workspace json file not exist: {self.path}")
        with open(self.path, 'r') as f:
            data = json.load(f)
            self.ws = CMDEditorWorkspace(raw_data=data)

        self._cfg_editors = {}

    def rename(self, new_name):
        assert not self.is_in_memory
        new_folder = os.path.join(Config.AAZ_DEV_WORKSPACE_FOLDER, new_name)
        if os.path.exists(new_folder):
            raise ValueError(f"Invalid new workspace folder: folder path exists: {new_folder}")
        os.rename(self.folder, new_folder)
        self.name = new_name
        self.folder = new_folder
        self.path = os.path.join(self.folder, 'ws.json')
        self.load()
        self.ws.name = new_name
        self.save()

    def delete(self):
        if not self.is_in_memory and os.path.exists(self.path):
            # make sure ws.json exist in folder
            if not os.path.isfile(self.path):
                raise exceptions.ResourceConflict(f"Workspace conflict: Is not file path: {self.path}")
            shutil.rmtree(self.folder)  # remove the whole folder
            return True
        return False

    def save(self):
        assert not self.is_in_memory

        if not os.path.exists(self.folder):
            os.makedirs(self.folder)

        remove_folders = []
        update_files = []
        used_resources = set()
        for resource_id, cfg_editor in self._cfg_editors.items():
            if resource_id in used_resources:
                continue
            for r_id, data in cfg_editor.iter_cfg_files_data():
                assert r_id not in used_resources
                if data is None:
                    remove_folders.append(WorkspaceCfgEditor.get_cfg_folder(self.folder, r_id))
                else:
                    update_files.append((WorkspaceCfgEditor.get_cfg_path(self.folder, r_id), data))
                used_resources.add(r_id)
        assert set(self._cfg_editors.keys()) == used_resources

        # verify ws timestamps
        # TODO: add write lock for path file
        if os.path.exists(self.path):
            with open(self.path, 'r') as f:
                data = json.load(f)
                pre_ws = CMDEditorWorkspace(data)
            if pre_ws.version != self.ws.version:
                raise exceptions.InvalidAPIUsage(f"Workspace Changed after: {self.ws.version}")

        self.ws.version = datetime.utcnow()
        with open(self.path, 'w') as f:
            data = json.dumps(self.ws.to_primitive(), ensure_ascii=False)
            f.write(data)

        for folder in remove_folders:
            shutil.rmtree(folder)

        for file_name, data in update_files:
            os.makedirs(os.path.dirname(file_name), exist_ok=True)
            with open(file_name, 'w') as f:
                f.write(data)

        self._cfg_editors = {}

    def find_command_tree_node(self, *node_names):
        node = self.ws.command_tree
        idx = 0
        while idx < len(node_names):
            name = node_names[idx]
            if not node.command_groups or name not in node.command_groups:
                return None
            node = node.command_groups[name]
            idx += 1
        return node

    def find_command_tree_leaf(self, *leaf_names):
        if len(leaf_names) < 1:
            raise exceptions.InvalidAPIUsage(f"Invalid command name: '{' '.join(leaf_names)}'")

        node = self.find_command_tree_node(*leaf_names[:-1])
        if not node:
            return None
        name = leaf_names[-1]
        if not node.commands or name not in node.commands:
            return None
        leaf = node.commands[name]
        return leaf

    def iter_command_tree_nodes(self, *root_node_names):
        """ Including the root node
        """
        root = self.find_command_tree_node(*root_node_names)
        if root:
            nodes = [root]  # add root node
            i = 0
            while i < len(nodes):
                yield nodes[i]
                for node in (nodes[i].command_groups or {}).values():
                    nodes.append(node)
                i += 1

    def iter_command_tree_leaves(self, *root_node_names):
        for node in self.iter_command_tree_nodes(*root_node_names):
            for leaf in (node.commands or {}).values():
                yield leaf

    def create_command_tree_nodes(self, *node_names):
        node = self.ws.command_tree
        idx = 0
        while idx < len(node_names):
            name = node_names[idx]
            if node.commands and name in node.commands:
                raise exceptions.InvalidAPIUsage(
                    f"Failed to create command group, '{' '.join(node_names[:idx+1])}' is command")
            if not node.command_groups or name not in node.command_groups:
                if not node.command_groups:
                    node.command_groups = {}
                aaz_node = self.aaz_specs.find_command_group(*node_names[:idx + 1])
                if aaz_node is not None:
                    new_node = CMDCommandTreeNode({
                        "names": node_names[:idx + 1],
                        "help": aaz_node.help.to_primitive()
                    })
                else:
                    new_node = CMDCommandTreeNode({
                        "names": node_names[:idx + 1],
                    })
                node.command_groups[name] = new_node
            node = node.command_groups[name]
            idx += 1
        return node

    def delete_command_tree_node(self, *node_names):
        for _ in self.iter_command_tree_leaves(*node_names):
            raise exceptions.ResourceConflict("Cannot delete command group with commands")
        parent = self.find_command_tree_node(*node_names[:-1])
        name = node_names[-1]
        if not parent or not parent.command_groups or name not in parent.command_groups:
            return False
        del parent.command_groups[name]
        if not parent.command_groups:
            parent.command_groups = None
        return True

    def check_resource_exist(self, resource_id, *root_node_names):
        for leaf in self.iter_command_tree_leaves(*root_node_names):
            for resource in leaf.resources:
                if resource.id == resource_id:
                    return True
        return False

    def add_cfg(self, cfg_editor, aaz_ref=None):
        cfg_editor.deleted = False
        for resource in cfg_editor.resources:
            self._cfg_editors[resource.id] = cfg_editor

        # update command tree
        for cmd_names, command in cfg_editor.iter_commands():
            node = self.create_command_tree_nodes(*cmd_names[:-1])
            name = cmd_names[-1]
            if node.commands is None:
                node.commands = {}
            assert name not in node.commands
            if node.command_groups:
                assert name not in node.command_groups
            reusable_leaf = self._reusable_leaves.pop(tuple(cmd_names), None)
            if reusable_leaf:
                new_cmd = reusable_leaf
            elif aaz_ref and (ref_v_name := aaz_ref.get(' '.join(cmd_names), None)) and (aaz_leaf := self.aaz_specs.find_command(*cmd_names)):
                # reference from aaz specs
                ref_v = None
                for v in aaz_leaf.versions:
                    if v.name == ref_v_name:
                        ref_v = v
                        break
                new_cmd = CMDCommandTreeLeaf({
                    "names": [*cmd_names],
                    "stage": ref_v.stage if ref_v else node.stage,
                    "help": aaz_leaf.help.to_primitive(),
                })
                if ref_v and ref_v.examples:
                    new_cmd.examples = []
                    for example in ref_v.examples:
                        new_cmd.examples.append(CMDCommandExample(example.to_primitive()))
            else:
                new_cmd = CMDCommandTreeLeaf({
                    "names": [*cmd_names],
                    "stage": node.stage,
                    "help": {
                        "short": command.description or ""
                    },
                })
            new_cmd.version = command.version
            new_cmd.resources = [CMDResource(r.to_primitive()) for r in command.resources]
            node.commands[name] = new_cmd

    def remove_cfg(self, cfg_editor):
        cfg_editor.deleted = True
        for resource in cfg_editor.resources:
            self._cfg_editors[resource.id] = cfg_editor

        # update command tree
        for cmd_names, _ in cfg_editor.iter_commands():
            node = self.find_command_tree_node(*cmd_names[:-1])
            name = cmd_names[-1]
            if node and node.commands and name in node.commands:
                # add into reusable leaves in case it's added in add_cfg again.
                self._reusable_leaves[tuple(cmd_names)] = node.commands.pop(name)

    def load_cfg_editor_by_resource(self, resource_id, version, reload=False):
        if not reload and resource_id in self._cfg_editors:
            # load from modified dict
            cfg_editor = self._cfg_editors[resource_id]
            return None if cfg_editor.deleted else cfg_editor
        assert not self.is_in_memory
        try:
            cfg_editor = WorkspaceCfgEditor.load_resource(self.folder, resource_id, version)
            for resource in cfg_editor.resources:
                self._cfg_editors[resource.id] = cfg_editor
            return cfg_editor
        except Exception as e:
            logger.error(f"load workspace resource cfg failed: {e}: {self.name} {resource_id} {version}")
            return None

    def load_cfg_editor_by_command(self, cmd, reload=False):
        return self.load_cfg_editor_by_resource(cmd.resources[0].id, cmd.resources[0].version, reload=reload)

    def update_command_tree_node_help(self, *node_names, help):
        node = self.find_command_tree_node(*node_names)
        if not node:
            raise exceptions.ResourceNotFind(f"Command Tree Node not found: '{' '.join(node_names)}'")

        if isinstance(help, CMDHelp):
            help = help.to_primitive()
        else:
            assert isinstance(help, dict)
        node.help = CMDHelp(help)
        return node

    def update_command_tree_leaf_help(self, *leaf_names, help):
        leaf = self.find_command_tree_leaf(*leaf_names)
        if not leaf:
            raise exceptions.ResourceNotFind(f"Command Tree leaf not found: '{' '.join(leaf_names)}'")

        if isinstance(help, CMDHelp):
            help = help.to_primitive()
        else:
            assert isinstance(help, dict)
        leaf.help = CMDHelp(help)
        return leaf

    def update_command_tree_node_stage(self, *node_names, stage):
        node = self.find_command_tree_node(*node_names)
        if not node:
            raise exceptions.ResourceNotFind(f"Command Tree Node not found: '{' '.join(node_names)}'")

        if node.stage == stage:
            return
        node.stage = stage

        if node.command_groups:
            for sub_node_name in node.command_groups:
                self.update_command_tree_node_stage(*node_names, sub_node_name, stage=stage)

        if node.commands:
            for leaf_name in node.commands:
                self.update_command_tree_leaf_stage(*node_names, leaf_name, stage=stage)
        return node

    def update_command_tree_leaf_stage(self, *leaf_names, stage):
        leaf = self.find_command_tree_leaf(*leaf_names)
        if not leaf:
            raise exceptions.ResourceNotFind(f"Command Tree leaf not found: '{' '.join(leaf_names)}'")

        if leaf.stage == stage:
            return
        leaf.stage = stage
        return leaf

    def update_command_tree_leaf_examples(self, *leaf_names, examples):
        leaf = self.find_command_tree_leaf(*leaf_names)
        if not leaf:
            raise exceptions.ResourceNotFind(f"Command Tree leaf not found: '{' '.join(leaf_names)}'")
        if not examples:
            leaf.examples = None
        else:
            leaf.examples = []
            for example in examples:
                example = CMDCommandExample(example)
                try:
                    example.validate()
                except Exception as err:
                # if not example.get('name', None) or not isinstance(example['name'], str):
                    raise exceptions.InvalidAPIUsage(f"Invalid example data: {err}")
                leaf.examples.append(example)
        return leaf

    def rename_command_tree_node(self, *node_names, new_node_names):
        new_name = ' '.join(new_node_names)
        if not new_name:
            raise exceptions.InvalidAPIUsage(f"Invalid new command name: {new_name}")

        node = self.find_command_tree_node(*node_names)
        if not node:
            raise exceptions.ResourceNotFind(f"Command Tree Node not found: '{' '.join(node_names)}'")

        if node.names == new_node_names:
            return

        parent = self.find_command_tree_node(*node.names[:-1])
        name = node.names[-1]
        if not parent or not parent.command_groups or name not in parent.command_groups or \
                node != parent.command_groups[name]:
            raise exceptions.ResourceConflict(f"Command Tree node not exist: '{' '.join(node.names)}'")

        self._pop_command_tree_node(parent, name)

        parent = self.create_command_tree_nodes(*new_node_names[:-1])
        return self._add_command_tree_node(parent, node, new_node_names[-1])

    def rename_command_tree_leaf(self, *leaf_names, new_leaf_names):
        new_name = ' '.join(new_leaf_names)
        if not new_name:
            raise exceptions.InvalidAPIUsage(f"Invalid new command name: {new_name}")

        leaf = self.find_command_tree_leaf(*leaf_names)
        if not leaf:
            raise exceptions.ResourceNotFind(f"Command Tree leaf not found: '{' '.join(leaf_names)}'")

        if leaf.names == new_leaf_names:
            return

        parent = self.find_command_tree_node(*leaf.names[:-1])
        name = leaf.names[-1]
        if not parent or not parent.commands or name not in parent.commands or leaf != parent.commands[name]:
            raise exceptions.ResourceConflict(f"Command Tree leaf not exist: '{' '.join(leaf.names)}")

        self._pop_command_tree_leaf(parent, name)

        parent = self.create_command_tree_nodes(*new_leaf_names[:-1])
        return self._add_command_tree_leaf(parent, leaf, new_leaf_names[-1])

    def generate_unique_name(self, *node_names, name):
        node = self.find_command_tree_node(*node_names)
        if not node:
            raise exceptions.ResourceConflict(f"Command Tree node not exist: '{' '.join(node_names)}'")
        if (not node.commands or name not in node.commands) and (
                not node.command_groups or name not in node.command_groups):
            return name
        idx = 1
        new_name = f"{name}-untitled{idx}"
        while node.commands and new_name in node.commands or node.command_groups and new_name in node.command_groups:
            idx += 1
            new_name = f"{name}-untitled{idx}"
        return new_name

    def add_new_resources_by_swagger(self, mod_names, version, resources):
        root_node = self.find_command_tree_node()
        assert root_node

        swagger_resources = []
        resource_options = []
        used_resource_ids = set()
        for r in resources:
            if r['id'] in used_resource_ids:
                continue
            if self.check_resource_exist(r['id']):
                raise exceptions.InvalidAPIUsage(f"Resource already added in Workspace: {r['id']}")
            # convert resource to swagger resource
            swagger_resource = self.swagger_specs.get_module_manager(
                plane=self.ws.plane, mod_names=mod_names
            ).get_resource_in_version(r['id'], version)
            swagger_resources.append(swagger_resource)
            resource_options.append(r.get("options", {}))
            used_resource_ids.update(r['id'])

        # load swagger resources
        self.swagger_command_generator.load_resources(swagger_resources)

        # generate cfg editors by resource
        cfg_editors = []
        aaz_ref = {}
        for resource, options in zip(swagger_resources, resource_options):
            try:
                command_group = self.swagger_command_generator.create_draft_command_group(resource, **options)
            except InvalidSwaggerValueError as err:
                raise exceptions.InvalidAPIUsage(
                    message=str(err)
                ) from err
            assert not command_group.command_groups, "The logic to support sub command groups is not supported"
            cfg_editor = WorkspaceCfgEditor.new_cfg(
                plane=self.ws.plane,
                resources=[resource.to_cmd()],
                command_groups=[command_group]
            )

            # inherit modification from cfg in aaz
            aaz_version = options.get('aaz_version', None)
            if aaz_version:
                try:
                    aaz_cfg_reader = self.aaz_specs.load_resource_cfg_reader(self.ws.plane, resource.id, aaz_version)
                except ValueError as err:
                    raise exceptions.InvalidAPIUsage(message=str(err)) from err
                cfg_editor.inherit_modification(aaz_cfg_reader)
                for cmd_names, _ in cfg_editor.iter_commands():
                    aaz_ref[' '.join(cmd_names)] = aaz_version

            cfg_editors.append(cfg_editor)

        # add cfg_editors
        self._add_cfg_editors(cfg_editors, aaz_ref=aaz_ref)

    def _add_cfg_editors(self, cfg_editors, aaz_ref=None):
        for cfg_editor in cfg_editors:
            merged = False
            rename_list = []
            for cmd_names, command in cfg_editor.iter_commands():
                cur_cmd = self.find_command_tree_leaf(*cmd_names)
                if cur_cmd is None:
                    continue
                if cur_cmd.version == command.version:
                    main_cfg_editor = self.load_cfg_editor_by_command(cur_cmd)
                    merged_cfg_editor = main_cfg_editor.merge(cfg_editor)
                    if merged_cfg_editor:
                        self.remove_cfg(main_cfg_editor)
                        self.add_cfg(merged_cfg_editor, aaz_ref=aaz_ref)
                        merged = True
                        break
                new_name = self.generate_unique_name(*cmd_names[:-1], name=cmd_names[-1])
                rename_list.append((cmd_names, [*cmd_names[:-1], new_name]))
            for cmd_names, new_cmd_names in rename_list:
                cfg_editor.rename_command(*cmd_names, new_cmd_names=new_cmd_names)
            if not merged:
                self.add_cfg(cfg_editor, aaz_ref=aaz_ref)

    def reload_swagger_resources(self, resources):
        reload_resource_map = {r['id']:{"version": r['version']} for r in resources}
        for leaf in self.iter_command_tree_leaves():
            ignore_resources = set()
            reload_versions = set()
            for r in leaf.resources:
                if r.id not in reload_resource_map:
                    ignore_resources.add(r.id)
                    continue
                reload_resource = reload_resource_map[r.id]
                version = reload_resource['version']
                reload_versions.add(version)
                if 'swagger_resource' not in reload_resource:
                    reload_resource['swagger_resource'] = self.swagger_specs.get_module_manager(
                        self.ws.plane, r.mod_names
                    ).get_resource_in_version(r.id, version)

                if 'cfg_editor' not in reload_resource:
                    reload_resource['cfg_editor'] = self.load_cfg_editor_by_command(leaf)
            if ignore_resources and len(ignore_resources) != len(leaf.resources):
                # not support partial resources reload
                raise exceptions.InvalidAPIUsage(f"Not support partial resources reload in one command: please select the following resources as well: {list(ignore_resources)}")
            if len(reload_versions) > 1:
                # not support multiple resource version for the same command
                raise exceptions.InvalidAPIUsage(f"Please select the same resource version for command: '{' '.join(leaf.names)}'")

        swagger_resources = []
        for resource_id, reload_resource in reload_resource_map.items():
            swagger_resource = reload_resource.get('swagger_resource', None)
            if not swagger_resource:
                raise exceptions.ResourceNotFind(f"Command not exist for '{resource_id}'")
            swagger_resources.append(swagger_resource)

        self.swagger_command_generator.load_resources(swagger_resources)

        new_cfg_editors = []
        for resource_id, reload_resource in reload_resource_map.items():
            options = {}
            cfg_editor = reload_resource['cfg_editor']
            swagger_resource = reload_resource['swagger_resource']
            methods = cfg_editor.get_used_http_methods(resource_id)
            if methods:
                options['methods'] = methods
            update_cmd_info = cfg_editor.get_update_cmd(resource_id)
            if update_cmd_info:
                _, _, update_by = update_cmd_info
                options['update_by'] = update_by
            try:
                command_group = self.swagger_command_generator.create_draft_command_group(
                    swagger_resource, **options)
            except InvalidSwaggerValueError as err:
                raise exceptions.InvalidAPIUsage(
                    message=str(err)
                ) from err
            assert not command_group.command_groups, "The logic to support sub command groups is not supported"
            new_cfg_editor = WorkspaceCfgEditor.new_cfg(
                plane=self.ws.plane,
                resources=[swagger_resource.to_cmd()],
                command_groups=[command_group]
            )
            new_cfg_editor.inherit_modification(cfg_editor)
            new_cfg_editors.append(new_cfg_editor)

        # remove old cfg editor
        for resource_id, reload_resource in reload_resource_map.items():
            cfg_editor = reload_resource['cfg_editor']
            self.remove_cfg(cfg_editor)

        # add cfg_editors
        self._add_cfg_editors(new_cfg_editors)

    def add_new_command_by_aaz(self, *cmd_names, version):
        # TODO: add support to load from aaz
        raise NotImplementedError()

    def _calculate_cfgs_common_command_group(self, cfg_editors, *node_names):
        # calculate common cg name prefix
        groups_names = []
        for cfg_editor in cfg_editors:
            for group in cfg_editor.cfg.command_groups:
                cg_names = group.name.split(" ")
                groups_names.append(cg_names)

        root_node = self.find_command_tree_node(*node_names)
        if len(node_names):
            # should also include the existing commands
            for leaf in (root_node.commands or {}).values():
                for leaf_resource in leaf.resources:
                    # cannot find match resource of resource_id with current mod_names and version
                    cg_names = self.swagger_command_generator.generate_command_group_name_by_resource(
                        resource_path=leaf_resource.swagger_path, rp_name=leaf_resource.rp_name)
                    cg_names = cg_names.split(" ")
                    groups_names.append(cg_names)

        common_prefix = groups_names[0]
        for names in groups_names[1:]:
            if len(names) < len(common_prefix):
                common_prefix = common_prefix[:len(names)]
            for i, k in enumerate(names):
                if i >= len(common_prefix):
                    break
                if common_prefix[i] != k:
                    common_prefix = common_prefix[:i]
                    break
        return common_prefix

    def get_resources(self, *root_node_names):
        resources = []
        used_resources = set()
        for leaf in self.iter_command_tree_leaves(*root_node_names):
            for resource in leaf.resources:
                if resource.id not in used_resources:
                    used_resources.add(resource.id)
                    resources.append(resource)
        resources = sorted(resources, key=lambda r: r.id)
        return resources

    def remove_resource(self, resource_id, version):
        cfg_editor = self.load_cfg_editor_by_resource(resource_id, version)
        if not cfg_editor:
            return False
        self.remove_cfg(cfg_editor)
        return True

    def list_commands_by_resource(self, resource_id, version):
        # will include all commands
        commands = []
        cfg_editor = self.load_cfg_editor_by_resource(resource_id, version)
        if cfg_editor:
            for cmd_names, _ in cfg_editor.iter_commands():
                leaf = self.find_command_tree_leaf(*cmd_names)
                if leaf:
                    commands.append(leaf)
        return commands

    def merge_resources(self, main_resource_id, main_resource_version, plus_resource_id, plus_resource_version):
        main_cfg_editor = self.load_cfg_editor_by_resource(main_resource_id, main_resource_version)
        plus_cfg_editor = self.load_cfg_editor_by_resource(plus_resource_id, plus_resource_version)
        merged_cfg_editor = main_cfg_editor.merge(plus_cfg_editor)
        if merged_cfg_editor:
            self.remove_cfg(plus_cfg_editor)
            self.remove_cfg(main_cfg_editor)
            self.add_cfg(merged_cfg_editor)
            return True
        return False

    def add_subresource_by_arg_var(self, resource_id, version, arg_var, cg_names, ref_args_options):

        cfg_editor = self.load_cfg_editor_by_resource(resource_id, version)
        if not cfg_editor:
            raise exceptions.InvalidAPIUsage(f"Resource not exist: resource_id={resource_id} version={version}")

        self.remove_cfg(cfg_editor)
        cfg_editor.build_subresource_commands_by_arg_var(resource_id, arg_var, cg_names, ref_args_options)
        self.add_cfg(cfg_editor)

    def remove_subresource(self, resource_id, version, subresource):
        cfg_editor = self.load_cfg_editor_by_resource(resource_id, version)
        if not cfg_editor:
            return False
        if not subresource:
            raise exceptions.InvalidAPIUsage(f"Invalid subresource: '{subresource}'")

        self.remove_cfg(cfg_editor)
        removed_commands = cfg_editor.remove_subresource_commands(resource_id, version, subresource)
        self.add_cfg(cfg_editor)
        return len(removed_commands) > 0

    def list_commands_by_subresource(self, resource_id, version, subresource):
        commands = []
        cfg_editor = self.load_cfg_editor_by_resource(resource_id, version)
        if cfg_editor:
            for cmd_names, _ in cfg_editor.iter_commands_by_resource(resource_id, subresource, version):
                leaf = self.find_command_tree_leaf(*cmd_names)
                if leaf:
                    commands.append(leaf)
        return commands

    @staticmethod
    def _pop_command_tree_node(parent, name):
        if not parent.command_groups or name not in parent.command_groups:
            raise IndexError(f"Command Tree node '{' '.join(parent.names)}' don't contain '{name}' sub node")
        return parent.command_groups.pop(name)

    @staticmethod
    def _pop_command_tree_leaf(parent, name):
        if not parent.commands or name not in parent.commands:
            raise IndexError(f"Command Tree node '{' '.join(parent.names)}' don't contain '{name}' leaf")
        return parent.commands.pop(name)

    def _add_command_tree_node(self, parent, node, name):
        command_groups = node.command_groups
        commands = node.commands

        node.command_groups = None
        node.commands = None

        # when it's conflict with command name, generate a unique name
        if parent.commands and name in parent.commands:
            new_name = self.generate_unique_name(*parent.names, name=name)
            logger.warning(f"Command Group name conflict with Command name: '{' '.join([*parent.names, name])}' : "
                           f"Use '{' '.join([*parent.names, new_name])}' instead")
            name = new_name

        if not parent.command_groups or name not in parent.command_groups:
            if not parent.command_groups:
                parent.command_groups = {}
            parent.command_groups[name] = node
            if parent == self.ws.command_tree:
                node.names = [name]
            else:
                node.names = [*parent.names, name]
        else:
            # merge with existing command group
            node = parent.command_groups[name]

        # add sub node and sub leaf
        if command_groups:
            for sub_name, sub_node in command_groups.items():
                self._add_command_tree_node(node, sub_node, sub_name)
        if commands:
            for sub_name, sub_leaf in commands.items():
                self._add_command_tree_leaf(node, sub_leaf, sub_name)
        return node

    def _add_command_tree_leaf(self, parent, leaf, name):
        cfg_editor = self.load_cfg_editor_by_command(leaf)

        # when it's conflict with command group name, generate a unique name
        if parent.command_groups and name in parent.command_groups:
            new_name = self.generate_unique_name(*parent.names, name=name)
            logger.warning(f"Command name conflict with Command Group name: '{' '.join([*parent.names, name])}' : "
                           f"Use '{' '.join([*parent.names, new_name])}' instead")
            name = new_name

        if parent.commands and name in parent.commands:
            assert leaf != parent.commands[name]
            new_name = self.generate_unique_name(*parent.names, name=name)
            logger.warning(f"Command name conflict with another Command's: '{' '.join([*parent.names, name])}' : "
                           f"Use '{' '.join([*parent.names, new_name])}' instead")
            name = new_name

        if not parent.commands:
            parent.commands = {}
        assert name not in parent.commands
        parent.commands[name] = leaf
        old_names = leaf.names
        if parent != self.ws.command_tree:
            new_cmd_names = [*parent.names, name]
        else:
            new_cmd_names = [name]
        leaf.names = [*new_cmd_names]
        cfg_editor.rename_command(*old_names, new_cmd_names=new_cmd_names)
        return leaf

    def generate_to_aaz(self):
        # Merge the commands of subresources which exported in aaz but not exist in current workspace
        self._merge_sub_resources_in_aaz()

        # update configurations
        for ws_leaf in self.iter_command_tree_leaves():
            editor = self.load_cfg_editor_by_command(ws_leaf)
            cfg = editor.cfg
            self.aaz_specs.update_resource_cfg(cfg)
        # update commands
        for ws_leaf in self.iter_command_tree_leaves():
            self.aaz_specs.update_command_by_ws(ws_leaf)
        # update command groups
        for ws_node in self.iter_command_tree_nodes():
            if ws_node == self.ws.command_tree:
                # ignore root node
                continue
            self.aaz_specs.update_command_group_by_ws(ws_node)
        self.aaz_specs.save()

    def _merge_sub_resources_in_aaz(self):
        """Merge the commands of subresources which exported in aaz but not exist in current workspace"""
        updated_cfgs = []
        inserted_commands = set()
        for ws_leaf in self.iter_command_tree_leaves():
            editor = self.load_cfg_editor_by_command(ws_leaf)
            existing_sub_resources = {}
            for cmd_names, command in editor.iter_commands():
                for r in command.resources:
                    if not r.subresource:
                        continue
                    key = (r.id, r.version)
                    if key not in existing_sub_resources:
                        existing_sub_resources[key] = set()
                    existing_sub_resources[key].add(r.subresource)
                    if r.subresource.endswith("[]") or r.subresource.endswith("{}"):
                        existing_sub_resources[key].add(r.subresource[:-2])

            aaz_ref = {}
            for (r_id, r_version), r_sub_resources in existing_sub_resources.items():
                pre_cfg_reader = self.aaz_specs.load_resource_cfg_reader(
                    editor.cfg.plane, resource_id=r_id, version=r_version
                )
                if not pre_cfg_reader:
                    continue

                for cmd_names, command in pre_cfg_reader.iter_commands():
                    insert_command = True
                    for r in command.resources:
                        if (r.id, r.version) not in existing_sub_resources:
                            insert_command = False
                            break
                        if not r.subresource or r.subresource in r_sub_resources:
                            insert_command = False
                            break
                    if not insert_command:
                        continue
                    cmd_names_str = ' '.join(cmd_names)
                    if cmd_names_str in inserted_commands:
                        continue
                    if self.find_command_tree_leaf(*cmd_names) is not None:
                        logger.error(
                            f"Command '{' '.join(cmd_names)}' in workspace conflict the name of Subresource Command in `aaz`")
                        continue
                    if self.find_command_tree_node(*cmd_names) is not None:
                        logger.error(
                            f"Command Group '{' '.join(cmd_names)}' in workspace conflict the name of Subresource Command in `aaz`")
                        continue
                    command = CMDCommand(command.to_native())
                    command.link()
                    editor._add_command(*cmd_names, command=command)
                    inserted_commands.add(cmd_names_str)
                    aaz_ref[cmd_names_str] = r_version
            if aaz_ref:
                editor.reformat()
                updated_cfgs = [(editor, aaz_ref)]

        for editor, aaz_ref in updated_cfgs:
            self.remove_cfg(editor)
            self.add_cfg(editor, aaz_ref)

    def find_similar_args(self, *cmd_names, arg):
        assert isinstance(arg, CMDArg)
        results = {}
        if arg.var.startswith("@"):
            # specify idx_suffix
            cls_name = arg.var[1:].replace('[', '.[').replace('{', '.{').split('.')[0]
            leaf = self.find_command_tree_leaf(*cmd_names)
            assert leaf is not None
            cfg_editor = self.load_cfg_editor_by_command(leaf)
            _, cls_arg, cls_arg_idx, _ = cfg_editor.find_arg_cls_definition(*cmd_names, cls_name=cls_name)
            _, arg_idx = cfg_editor.find_arg_by_var(*cmd_names, arg_var=arg.var)
            assert arg_idx.startswith(cls_arg_idx)
            idx_suffix = arg_idx[len(cls_arg_idx):]
            assert len(idx_suffix) > 0

            cls_name_prefix = cls_name.split('_')[0]  # remove the subfix such as `_create` `_update`
            for leaf in self.iter_command_tree_leaves():
                cfg_editor = self.load_cfg_editor_by_command(leaf)
                for _, similar_cls_arg, similar_cls_arg_idx, _ in cfg_editor.iter_arg_cls_definition(
                        *leaf.names, cls_name_prefix=cls_name_prefix):
                    # search cls definition in command
                    # find sub arg by idx_suffix
                    similar_arg = cfg_editor.find_sub_arg(similar_cls_arg, idx=idx_suffix)
                    if similar_arg is None or not cfg_editor.is_similar_args(arg, similar_arg):
                        continue
                    similar_arg_idx = similar_cls_arg_idx + idx_suffix

                    key = tuple(leaf.names)
                    assert key not in results
                    results[key] = {
                        similar_arg.var: [similar_arg_idx]
                    }

                    # search cls reference in command
                    for _, _, ref_arg_idx, _ in cfg_editor.iter_arg_cls_reference(*leaf.names, cls_name=similar_cls_arg.cls):
                        results[key][similar_arg.var].append(ref_arg_idx + idx_suffix)

        else:
            for leaf in self.iter_command_tree_leaves():
                cfg_editor = self.load_cfg_editor_by_command(leaf)
                similar_arg, similar_arg_idx = cfg_editor.find_arg_by_var(*leaf.names, arg_var=arg.var)
                if similar_arg is None or not cfg_editor.is_similar_args(arg, similar_arg):
                    continue
                key = tuple(leaf.names)
                assert key not in results
                results[key] = {
                    similar_arg.var: [similar_arg_idx]
                }
        return results
